import contextlib as _contextlib
import datetime as _dt
import logging as _logging
import re as _re
import textwrap as _textwrap
from collections import namedtuple
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
)

import sqlalchemy as _sa
from sqlalchemy_utils import UUIDType as _UUIDType

from ._compat import PsycoPgLockNotAvailable, SAConnection, SARow
from ._factories import SubscriptionFactory
from ._interfaces import (
    AggregatedStreamMessage,
    MessagePartitioner,
    MessageProtocol,
    RunOnNotificationResult,
    StoredMessage,
    StreamPartitionStatistic,
    SubscriptionStartPoint,
    TimeBudget,
)
from ._message_store import MessageStore

if TYPE_CHECKING:
    from ._aggregated_stream_reader import (
        AggregatedStreamReader,
        AsyncAggregatedStreamReader,
    )

E = TypeVar("E", bound=MessageProtocol)

LOGGER = _logging.getLogger(__name__)


class AggregatedStream(Generic[E]):
    def __init__(
        self,
        name: str,
        store: MessageStore[E],
        partitioner: MessagePartitioner[E],
        stream_wildcards: List[str],
        update_batch_size: Optional[int] = None,
        lookback_for_gaps_hours: Optional[int] = None,
    ) -> None:
        """
        AggregatedStream aggregates multiple streams into one (partitioned) stream.

        Read more about aggregated streams under [Data Model](../concepts/data-model.md#aggregated-streams).

        The `update_batch_size` argument can be used to control the batch size of the
        update process. Higher numbers will result in less database roundtrips but
        also in higher memory usage.

        Args:
            name: Stream name, needs to be a valid python identifier
            store: Message store
            partitioner: Message partitioner
            stream_wildcards: List of stream wildcards
            update_batch_size: Batch size for updating the stream, defaults to 100
            lookback_for_gaps_hours: How many hours we should look aback for gaps in global positions. (Default is 6 hours, set this to 2-4x the time your longest transaction takes)

        Attributes:
            name (str): Stream name
            projector (StreamProjector): Stream projector
            subscription (SubscriptionFactory): Factory to create subscriptions on this stream
        """
        assert name.isidentifier(), "name must be a valid identifier"
        self.name = name
        self.subscription = SubscriptionFactory(self)
        self._store = store
        self._metadata = _sa.MetaData()
        self._table = _sa.Table(
            self.stream_table_name(name),
            self._metadata,
            _sa.Column("message_id", _UUIDType(), primary_key=True),
            _sa.Column("origin_stream", _sa.String(255), nullable=False),
            _sa.Column("origin_stream_version", _sa.Integer, nullable=False),
            _sa.Column(
                "partition",
                _sa.Integer,
                nullable=False,
                index=True,
            ),
            _sa.Column(
                "position",
                _sa.Integer,
                nullable=False,
            ),
            _sa.Column(
                "message_occurred_at",
                _sa.DateTime,
                nullable=False,
            ),
            _sa.Column(
                "origin_stream_global_position", _sa.Integer, nullable=False, index=True
            ),
            _sa.Column("origin_stream_added_at", _sa.DateTime, nullable=False),
            _sa.UniqueConstraint(
                "partition",
                "position",
                name=f"depeche_stream_{name}_uq",
            ),
        )
        self.notification_channel = self.notification_channel_name(name)
        trigger = _sa.DDL(
            _notify_trigger(
                name=name,
                tablename=self._table.name,
                notification_channel=self.notification_channel,
            )
        )
        _sa.event.listen(
            self._table, "after_create", trigger.execute_if(dialect="postgresql")
        )
        self._metadata.create_all(store.engine, checkfirst=True)
        self.projector = StreamProjector(
            stream=self,
            partitioner=partitioner,
            stream_wildcards=stream_wildcards,
            batch_size=update_batch_size,
            lookback_for_gaps_hours=lookback_for_gaps_hours,
        )

    def truncate(self, conn: SAConnection):
        """
        Truncate aggregated stream.
        """
        conn.execute(self._table.delete())

    def _add(
        self,
        conn: SAConnection,
        rows,
    ) -> None:
        conn.execute(self._table.insert().values(rows))

    def read(
        self, partition: int, conn: Optional[SAConnection] = None
    ) -> Iterator[AggregatedStreamMessage]:
        """
        Read all messages from a partition of the aggregated stream.

        Args:
            partition: Partition number
            conn: Optional connection to use for reading. If not provided, a new connection will be created.
        """

        def _inner(conn):
            for row in conn.execute(
                _sa.select(self._table.c.message_id, self._table.c.position)
                .where(self._table.c.partition == partition)
                .order_by(self._table.c.position)
            ):
                yield AggregatedStreamMessage(
                    message_id=row.message_id,
                    position=row.position,
                    partition=partition,
                )

        if conn is None:
            with self._connection() as conn:
                yield from _inner(conn)
        else:
            yield from _inner(conn)

    def read_slice(
        self,
        partition: int,
        start: int,
        count: int,
        conn: Optional[SAConnection] = None,
    ) -> Iterator[AggregatedStreamMessage]:
        """
        Read a slice of messages from a partition of the aggregated stream.

        Args:
            partition: Partition number
            start: Start position
            count: Number of messages to read
            conn: Optional connection to use for reading. If not provided, a new connection will be created.
        """

        def _inner(conn):
            for row in conn.execute(
                _sa.select(self._table.c.message_id, self._table.c.position)
                .where(
                    _sa.and_(
                        self._table.c.partition == partition,
                        self._table.c.position >= start,
                    )
                )
                .order_by(self._table.c.position)
                .limit(count)
            ):
                yield AggregatedStreamMessage(
                    message_id=row.message_id,
                    position=row.position,
                    partition=partition,
                )

        if conn is None:
            with self._connection() as conn:
                yield from _inner(conn)
        else:
            yield from _inner(conn)

    def reader(
        self, start_point: Optional[SubscriptionStartPoint] = None
    ) -> "AggregatedStreamReader":
        """
        Get a reader for the aggregated stream.

        Args:
            start_point: Start point for the reader
        """
        from ._aggregated_stream_reader import AggregatedStreamReader

        return AggregatedStreamReader(self, start_point=start_point)

    def async_reader(
        self, start_point: Optional[SubscriptionStartPoint] = None
    ) -> "AsyncAggregatedStreamReader":
        """
        Get an async reader for the aggregated stream.

        Args:
            start_point: Start point for the reader
        """
        from ._aggregated_stream_reader import AsyncAggregatedStreamReader

        return AsyncAggregatedStreamReader(self, start_point=start_point)

    @_contextlib.contextmanager
    def _connection(self):
        conn = self._store.engine.connect()
        try:
            yield conn
        finally:
            conn.close()

    def get_partition_statistics(
        self,
        result_limit: Optional[int] = None,
        conn: Optional[SAConnection] = None,
    ) -> Iterator[StreamPartitionStatistic]:
        def _inner(conn):
            tbl = self._table.alias()
            next_messages_tbl = (
                _sa.select(
                    tbl.c.partition,
                    _sa.func.min(tbl.c.position).label("min_position"),
                    _sa.func.max(tbl.c.position).label("max_position"),
                )
                .group_by(tbl.c.partition)
                .cte()
            )

            qry = _sa.select(tbl, next_messages_tbl.c.max_position).select_from(
                next_messages_tbl.join(
                    tbl,
                    _sa.and_(
                        tbl.c.partition == next_messages_tbl.c.partition,
                        tbl.c.position == next_messages_tbl.c.min_position,
                    ),
                )
            )
            if result_limit is not None:
                qry = qry.order_by(tbl.c.message_occurred_at).limit(result_limit)
            result = conn.execute(qry)
            for row in result.fetchall():
                yield StreamPartitionStatistic(
                    partition_number=row.partition,
                    next_message_id=row.message_id,
                    next_message_position=row.position,
                    next_message_occurred_at=row.message_occurred_at,
                    max_position=row.max_position,
                )
            result.close()
            del result

        if conn is None:
            with self._connection() as conn:
                yield from _inner(conn)
        else:
            yield from _inner(conn)

    def _get_max_aggregated_stream_positions(
        self,
        conn: SAConnection,
        # min_position: int,
    ) -> Dict[int, int]:
        # Relatively expensive operation, so we should try hard to do it only when required
        tbl = self._table
        qry = (
            _sa.select(
                tbl.c.partition,
                _sa.func.max(tbl.c.position).label("max_position"),
            )
            # TODO this filter would be very helpful performance-wise!
            # .where(tbl.c.position >= min_position)
            .group_by(tbl.c.partition)
        )
        result = {
            row.partition: row.max_position for row in conn.execute(qry).fetchall()
        }
        return result

    def time_to_positions(self, time: _dt.datetime) -> Dict[int, int]:
        """
        Get the positions for each partition at a given time.

        Args:
            time: Time to get positions for (must be timezone aware)

        Returns:
            A dictionary mapping partition numbers to positions
        """
        if time.tzinfo is None:
            raise ValueError("time must be timezone aware")
        with self._connection() as conn:
            tbl = self._table.alias()

            positions_after_time = (
                _sa.select(
                    tbl.c.partition,
                    _sa.func.min(tbl.c.position).label("min_position"),
                )
                .where(tbl.c.message_occurred_at >= time.astimezone(_dt.timezone.utc))
                .group_by(tbl.c.partition)
                .cte()
            )
            max_positions = (
                _sa.select(
                    tbl.c.partition,
                    _sa.func.max(tbl.c.position).label("max_position"),
                )
                .group_by(tbl.c.partition)
                .cte()
            )
            qry = _sa.select(
                max_positions.c.partition,
                _sa.func.coalesce(
                    positions_after_time.c.min_position,
                    max_positions.c.max_position + 1,
                ).label("position"),
            ).select_from(
                max_positions.outerjoin(
                    positions_after_time,
                    max_positions.c.partition == positions_after_time.c.partition,
                )
            )
            return {row.partition: row.position for row in conn.execute(qry)}

    def global_position_to_positions(self, global_position: int) -> Dict[int, int]:
        """
        Get the positions for each partition at a given global position.

        Args:
            global_position: Global position

        Returns:
            A dictionary mapping partition numbers to positions
        """
        with self._connection() as conn:
            tbl = self._table.alias()
            messages_tbl = self._store._storage.message_table.alias()

            positions_upto_global_pos = (
                _sa.select(
                    tbl.c.partition,
                    _sa.func.max(tbl.c.position).label("position"),
                )
                .select_from(
                    tbl.join(
                        messages_tbl,
                        messages_tbl.c.message_id == tbl.c.message_id,
                    )
                )
                .where(messages_tbl.c.global_position <= global_position)
                .group_by(tbl.c.partition)
                .cte()
            )
            partitions = (
                _sa.select(
                    tbl.c.partition,
                )
                .group_by(tbl.c.partition)
                .cte()
            )
            qry = _sa.select(
                partitions.c.partition,
                _sa.func.coalesce(
                    positions_upto_global_pos.c.position,
                    -1,
                ).label("position"),
            ).select_from(
                partitions.outerjoin(
                    positions_upto_global_pos,
                    partitions.c.partition == positions_upto_global_pos.c.partition,
                )
            )
            return {row.partition: row.position for row in conn.execute(qry)}

    @staticmethod
    def stream_table_name(name: str) -> str:
        return f"depeche_stream_{name}"

    @staticmethod
    def notification_channel_name(name: str) -> str:
        return f"depeche_{name}_messages"

    @classmethod
    def get_migration_ddl_0_8_0(
        cls, aggregated_stream_name: str, message_store_name: str
    ) -> str:
        """
        DDL Script to migrate from < 0.8.0
        """
        del message_store_name  # Unused in this migration
        tablename = cls.stream_table_name(aggregated_stream_name)
        new_objects = _notify_trigger(
            name=aggregated_stream_name,
            tablename=tablename,
            notification_channel=cls.notification_channel_name(aggregated_stream_name),
        )
        return _textwrap.dedent(
            f"""
            ALTER TABLE "{aggregated_stream_name}_projected_stream"
                 RENAME TO {tablename};
            DROP TRIGGER IF EXISTS {aggregated_stream_name}_stream_notify_message_inserted;
            DROP FUNCTION IF EXISTS {aggregated_stream_name}_stream_notify_message_inserted;
            {new_objects}
            """
        )

    @classmethod
    def get_migration_ddl_0_11_0(
        cls, aggregated_stream_name: str, message_store_name: str
    ) -> str:
        """
        DDL Script to migrate from < 0.11.0
        """
        aggregated_stream_tablename = cls.stream_table_name(aggregated_stream_name)
        message_tablename = f"depeche_msgs_{message_store_name}"
        return _textwrap.dedent(
            f"""
            -- Add columns (nullable)
            ALTER TABLE {aggregated_stream_tablename}
                ADD COLUMN origin_stream_global_position INTEGER NULL,
                ADD COLUMN origin_stream_added_at TIMESTAMP NULL;

            -- Copy data from the message store to the aggregated stream
            UPDATE {aggregated_stream_tablename} AS agg
                SET origin_stream_global_position = msg.global_position,
                origin_stream_added_at = msg.added_at
                FROM {message_tablename} AS msg
                WHERE agg.message_id = msg.message_id;

            -- Make the new columns NOT NULL
            ALTER TABLE {aggregated_stream_tablename}
                ALTER COLUMN origin_stream_global_position SET NOT NULL,
                ALTER COLUMN origin_stream_added_at SET NOT NULL;

            --- Add index
            CREATE INDEX ix_{aggregated_stream_tablename}_origin_stream_global_position
                ON {aggregated_stream_tablename} (origin_stream_global_position);
            """
        )

    @classmethod
    def migration_script_generators(
        cls,
    ) -> Dict[Tuple[int, int], List[Callable[[str, str], str]]]:
        return {
            (0, 8): [
                cls.get_migration_ddl_0_8_0,
            ],
            (0, 11): [
                cls.get_migration_ddl_0_11_0,
            ],
        }


def _notify_trigger(name: str, tablename: str, notification_channel: str) -> str:
    trigger_name = f"depeche_stream_new_msg_{name}"
    return f"""
        CREATE OR REPLACE FUNCTION {trigger_name}()
          RETURNS trigger AS $$
        DECLARE
        BEGIN
          PERFORM pg_notify(
            '{notification_channel}',
            json_build_object(
                'message_id', NEW.message_id,
                'partition', NEW.partition,
                'position', NEW.position
            )::text);
          RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;

        CREATE TRIGGER {trigger_name}
          AFTER INSERT ON {tablename}
          FOR EACH ROW
          EXECUTE PROCEDURE {trigger_name}();
        """


class _AlreadyUpdating(RuntimeError):
    pass


SelectedOriginStream = namedtuple(
    "SelectedOriginStream",
    [
        "stream",
        "start_at_global_position",
        # "estimated_message_count",
    ],
)

AggregatedStreamPositon = namedtuple(
    "AggregatedStreamPositon",
    [
        "origin_stream",
        "max_aggregated_stream_global_position",
        "min_aggregated_stream_global_position",
    ],
)

OriginStreamPositon = namedtuple(
    "OriginStreamPositon",
    [
        "origin_stream",
        "max_global_position",
        "min_global_position",
    ],
)

AggregatedStreamHead = namedtuple(
    "AggregatedStreamHead",
    [
        "global_position",
        "added_at",
    ],
)

LookbackCache = namedtuple(
    "LookbackCache",
    [
        "head_added_at",
        "value",
    ],
)

OriginStreamPositionsCache = namedtuple(
    "OriginStreamPositionsCache",
    [
        "estimated_gap_look_back_start",
        "origin_streams",
    ],
)

FullUpdateResult = namedtuple(
    "FullUpdateResult",
    [
        "n_updated_messages",
        "more_messages_available",
    ],
)


class StreamProjector(Generic[E]):
    def __init__(
        self,
        stream: AggregatedStream[E],
        partitioner: MessagePartitioner[E],
        stream_wildcards: List[str],
        batch_size: Optional[int] = None,
        lookback_for_gaps_hours: Optional[int] = None,
    ):
        """
        Stream projector is responsible for updating an aggregated stream.

        The update process is locked to prevent concurrent updates. Thus, it is
        fine to run the projector in multiple processes.

        Implements: [RunOnNotification][depeche_db.RunOnNotification]
        """
        self.stream = stream
        self.stream_wildcards = stream_wildcards
        self.stream_regexes = [
            _re.compile(
                wildcard.replace("\\", "\\\\").replace(".", "\\.").replace("%", ".*")
            )
            for wildcard in stream_wildcards
        ]
        self.partitioner = partitioner
        self.batch_size = batch_size or 100
        self.lookback_for_gaps_hours = lookback_for_gaps_hours or 6
        self._lookback_cache: Optional[LookbackCache] = None
        self._get_origin_stream_positions_cache: Optional[
            OriginStreamPositionsCache
        ] = None

    def interested_in_notification(self, notification: dict) -> bool:
        # Check if the projector is interested in the notification.
        stream = notification.get("stream")
        if isinstance(stream, str):
            for regex in self.stream_regexes:
                if regex.fullmatch(stream):
                    return True
        return False

    def take_notification_hint(self, notification: dict):
        pass

    @property
    def notification_channel(self) -> str:
        """
        Returns the notification channel name for this projector.
        """
        return self.stream._store._storage.notification_channel

    def run(self, budget: Optional[TimeBudget] = None) -> RunOnNotificationResult:
        """
        Runs the projector once.
        """
        try:
            result = self.update_full(budget=budget)
            if budget and budget.over_budget() and result.more_messages_available:
                return RunOnNotificationResult.WORK_REMAINING
        except _AlreadyUpdating:
            pass
        return RunOnNotificationResult.DONE_FOR_NOW

    def stop(self):
        """
        No-Op on this class.
        """
        pass

    def update_full(self, budget: Optional[TimeBudget] = None) -> FullUpdateResult:
        """
        Updates the projection from the last known position to the current position.
        """
        result = 0
        batch_num = 0
        with self.stream._store.engine.connect() as conn:
            cutoff = self.stream._store._storage.get_global_position(conn)
            try:
                conn.execute(
                    _sa.text(
                        f"LOCK TABLE {self.stream._table.name} IN EXCLUSIVE MODE NOWAIT"
                    )
                )
            except _sa.exc.OperationalError as exc:
                if isinstance(exc.orig, PsycoPgLockNotAvailable):
                    self._aggregate_stream_positions_cache = None
                    raise _AlreadyUpdating(
                        "Cannot update stream projection, because another process is already updating it."
                    )
                raise
            while True:
                batch_num = self._update_batch(conn, cutoff)
                result += batch_num
                LOGGER.debug(f"{self.stream.name}: Batch updated: {batch_num} messages")
                if batch_num < self.batch_size:
                    # No more messages to process
                    break
                if budget and budget.over_budget():
                    # Budget exceeded, stop processing
                    break
            conn.commit()
        return FullUpdateResult(result, batch_num == self.batch_size)

    def get_aggregate_stream_positions(
        self,
        conn: SAConnection,
        estimated_gap_look_back_start: int,
    ) -> Dict[str, AggregatedStreamPositon]:
        stream_table = self.stream._table.alias()
        return {
            row.origin_stream: AggregatedStreamPositon(
                origin_stream=row.origin_stream,
                max_aggregated_stream_global_position=row.max_aggregated_stream_global_position,
                min_aggregated_stream_global_position=row.min_aggregated_stream_global_position,
            )
            for row in conn.execute(
                _sa.select(
                    stream_table.c.origin_stream,
                    _sa.func.max(stream_table.c.origin_stream_global_position).label(
                        "max_aggregated_stream_global_position"
                    ),
                    _sa.func.min(stream_table.c.origin_stream_global_position).label(
                        "min_aggregated_stream_global_position"
                    ),
                )
                .where(
                    stream_table.c.origin_stream_global_position
                    >= estimated_gap_look_back_start,
                )
                .group_by(stream_table.c.origin_stream),
            )
        }

    def get_origin_stream_positions(
        self,
        conn: SAConnection,
        min_global_position: int,
        max_global_position: Optional[int] = None,
    ) -> Dict[str, OriginStreamPositon]:
        origin_table = self.stream._store._storage.message_table
        cutoff_cond = []
        if max_global_position is not None:
            cutoff_cond = [origin_table.c.global_position <= max_global_position]
        return {
            stream: OriginStreamPositon(
                origin_stream=stream,
                max_global_position=max_global_position,
                min_global_position=min_global_position,
            )
            for (
                stream,
                max_global_position,
                min_global_position,
            ) in conn.execute(
                # TODO execute this with pyscopg's `prepare` option False
                # https://www.psycopg.org/psycopg3/docs/api/connections.html#psycopg.Connection.execute
                _sa.select(
                    origin_table.c.stream,
                    _sa.func.max(origin_table.c.global_position).label(
                        "max_origin_stream_global_position"
                    ),
                    _sa.func.min(origin_table.c.global_position).label(
                        "min_origin_stream_global_position"
                    ),
                )
                .where(
                    _sa.and_(
                        origin_table.c.global_position >= min_global_position,
                        _sa.or_(
                            *[
                                origin_table.c.stream.like(wildcard)
                                for wildcard in self.stream_wildcards
                            ]
                        ),
                        *cutoff_cond,
                    )
                )
                .group_by(origin_table.c.stream),
            ).fetchall()
        }

    def _get_aggregated_stream_head(self, conn: SAConnection) -> AggregatedStreamHead:
        stream_table = self.stream._table.alias()
        row = conn.execute(
            _sa.select(
                stream_table.c.origin_stream_global_position,
                stream_table.c.origin_stream_added_at,
            )
            .order_by(stream_table.c.origin_stream_global_position.desc())
            .limit(1)
        ).fetchone()
        if row:
            head_global_position, head_added_at = row
            return AggregatedStreamHead(
                global_position=head_global_position, added_at=head_added_at
            )
        return AggregatedStreamHead(
            global_position=-1,
            added_at=_dt.datetime(1980, 1, 1, tzinfo=_dt.timezone.utc),
        )

    def _estimate_gap_look_back_start(
        self, conn: SAConnection, head_added_at: _dt.datetime
    ) -> int:
        if self._lookback_cache is not None:
            if (head_added_at - self._lookback_cache.head_added_at) > _dt.timedelta(
                hours=1
            ):
                self._lookback_cache = None

        if self._lookback_cache is None:
            LOGGER.debug(f"{self.stream.name}: Updating lookback estimation")
            origin_table = self.stream._store._storage.message_table
            value = (
                conn.execute(
                    _sa.select(_sa.func.max(origin_table.c.global_position)).where(
                        origin_table.c.added_at
                        <= head_added_at
                        - _dt.timedelta(hours=self.lookback_for_gaps_hours)
                    )
                ).scalar_one_or_none()
            ) or 0
            self._lookback_cache = LookbackCache(
                head_added_at=head_added_at, value=value
            )

        return self._lookback_cache.value  # type: ignore

    def _select_origin_streams(
        self, conn: SAConnection, cutoff: Optional[int] = None
    ) -> List[SelectedOriginStream]:
        head = self._get_aggregated_stream_head(conn)
        if head.global_position > 0:
            estimated_gap_look_back_start = self._estimate_gap_look_back_start(
                conn, head.added_at
            )
        else:
            estimated_gap_look_back_start = 0
        LOGGER.debug(
            f"{self.stream.name}: Estimated gap look back start: {estimated_gap_look_back_start}"
        )

        stream_positions = self.get_aggregate_stream_positions(
            conn,
            estimated_gap_look_back_start=estimated_gap_look_back_start,
        )

        candidate_streams: List[SelectedOriginStream] = []
        if self._get_origin_stream_positions_cache is not None:
            if (
                self._get_origin_stream_positions_cache.estimated_gap_look_back_start
                == estimated_gap_look_back_start
            ):
                # We are using the cached origin streams
                candidate_streams = self._calculate_selected_streams(
                    origin_streams=self._get_origin_stream_positions_cache.origin_streams,
                    stream_positions=stream_positions,
                    estimated_gap_look_back_start=estimated_gap_look_back_start,
                )
                if candidate_streams:
                    LOGGER.debug(
                        f"{self.stream.name}: Found {len(candidate_streams)} candidate streams (using cached origin streams)"
                    )
        if not candidate_streams:
            # Cached origin streams either did not give us any candidates or
            # were not available.
            origin_streams = self.get_origin_stream_positions(
                conn=conn,
                min_global_position=estimated_gap_look_back_start,
                max_global_position=cutoff,
            )
            self._get_origin_stream_positions_cache = OriginStreamPositionsCache(
                estimated_gap_look_back_start=estimated_gap_look_back_start,
                origin_streams=origin_streams,
            )
            candidate_streams = self._calculate_selected_streams(
                origin_streams=origin_streams,
                stream_positions=stream_positions,
                estimated_gap_look_back_start=estimated_gap_look_back_start,
            )
            LOGGER.debug(
                f"{self.stream.name}: Found {len(candidate_streams)} candidate streams (using live origin streams)"
            )

        # TODO limit so that sum(estimated_message_count) close to batch_size
        return sorted(candidate_streams, key=lambda x: x.start_at_global_position)[
            : self.batch_size
        ]

    def _calculate_selected_streams(
        self,
        origin_streams: Dict[str, OriginStreamPositon],
        stream_positions: Dict[str, AggregatedStreamPositon],
        estimated_gap_look_back_start: int,
    ) -> List[SelectedOriginStream]:
        candidate_streams: List[SelectedOriginStream] = []
        for origin_stream in origin_streams.values():
            stream_position = stream_positions.get(origin_stream.origin_stream)
            if stream_position is None:
                assert (
                    origin_stream.min_global_position >= estimated_gap_look_back_start
                )
                candidate_streams.append(
                    SelectedOriginStream(
                        stream=origin_stream.origin_stream,
                        start_at_global_position=origin_stream.min_global_position,
                        # TODO add estimate? or exact count?
                        # estimated_message_count=origin_stream.max_version,
                    )
                )
            elif (
                stream_position.max_aggregated_stream_global_position
                < origin_stream.max_global_position
            ):
                candidate_streams.append(
                    SelectedOriginStream(
                        stream=origin_stream.origin_stream,
                        start_at_global_position=(
                            stream_position.max_aggregated_stream_global_position + 1
                        ),
                        # TODO add estimate? or exact count?
                        # estimated_message_count=origin_stream.max_version - stream_position.max_aggregated_stream_version,
                    )
                )
        return candidate_streams

    def _update_batch(self, conn: SAConnection, cutoff: Optional[int] = None) -> int:
        message_table = self.stream._store._storage.message_table

        selected_origin_streams = self._select_origin_streams(conn, cutoff=cutoff)
        if not selected_origin_streams:
            return 0

        min_global_position = min(
            selected_stream.start_at_global_position
            for selected_stream in selected_origin_streams
        )
        cutoff_cond = []
        if cutoff is not None:
            cutoff_cond = [message_table.c.global_position <= cutoff]
        qry = (
            _sa.select(
                message_table.c.message_id,
                message_table.c.stream,
                message_table.c.version,
                message_table.c.message,
                message_table.c.global_position,
                message_table.c.added_at,
            )
            .where(
                _sa.and_(
                    message_table.c.global_position >= min_global_position,
                    *cutoff_cond,
                    _sa.or_(
                        *[
                            _sa.and_(
                                message_table.c.stream == selected_stream.stream,
                                message_table.c.global_position
                                >= selected_stream.start_at_global_position,
                            )
                            for selected_stream in selected_origin_streams
                        ]
                    ),
                )
            )
            .order_by(message_table.c.global_position)
            .limit(self.batch_size)
        )
        messages = list(conn.execute(qry).fetchall())

        LOGGER.debug(f"{self.stream.name}: Found {len(messages)} new messages")
        if not messages:
            return 0

        self._add(conn, messages)
        return len(messages)

    def _add(self, conn: SAConnection, messages: List[SARow]) -> None:
        positions = {
            row.partition: row.max_position
            for row in conn.execute(
                _sa.select(
                    self.stream._table.c.partition,
                    _sa.func.max(self.stream._table.c.position).label("max_position"),
                ).group_by(self.stream._table.c.partition)
            )
        }

        rows = []
        for message_id, stream, version, message, global_position, added_at in messages:
            message = StoredMessage(
                message_id=message_id,
                stream=stream,
                version=version,
                message=self.stream._store._serializer.deserialize(message),
                global_position=global_position,
                added_at=added_at,
            )
            partition = self.partitioner.get_partition(message)
            if partition < 0:
                raise ValueError("partition must be >= 0")
            position = positions.get(partition, -1) + 1
            positions[partition] = position
            rows.append(
                (
                    message_id,
                    stream,
                    version,
                    partition,
                    position,
                    message.message.get_message_time(),
                    global_position,
                    added_at,
                )
            )
        self.stream._add(
            conn=conn,
            rows=rows,
        )



from ._aggregated_stream import (  # noqa: F401
    AggregatedStream,
    StreamProjector,
)
from ._aggregated_stream_reader import (  # noqa: F401
    AggregatedStreamReader,
    AsyncAggregatedStreamReader,
)
from ._exceptions import (  # noqa: F401
    MessageIdMismatchError,
    MessageNotFound,
    OptimisticConcurrencyError,
)
from ._executor import Executor  # noqa: F401
from ._factories import AggregatedStreamFactory, SubscriptionFactory  # noqa: F401
from ._interfaces import (  # noqa: F401
    CallMiddleware,
    ErrorAction,
    FixedTimeBudget,
    HandlerDescriptor,
    LockProvider,
    MessageHandlerRegisterProtocol,
    MessagePartitioner,
    MessagePosition,
    MessageProtocol,
    MessageSerializer,
    RunOnNotification,
    RunOnNotificationResult,
    StoredMessage,
    StreamPartitionStatistic,
    SubscriptionErrorHandler,
    SubscriptionMessage,
    SubscriptionStartPoint,
    SubscriptionState,
    SubscriptionStateProvider,
    TimeBudget,
)
from ._message_handler import MessageHandler, MessageHandlerRegister  # noqa: F401
from ._message_store import MessageStore, MessageStoreReader  # noqa: F401
from ._storage import Storage  # noqa: F401
from ._subscription import (  # noqa: F401
    AckStrategy,
    BatchedAckSubscriptionRunner,
    ExitSubscriptionErrorHandler,  # move somewhere else
    LogAndIgnoreSubscriptionErrorHandler,  # move somewhere else
    StartAtNextMessage,
    StartAtPointInTime,
    Subscription,
    SubscriptionMessageHandler,
    SubscriptionRunner,
)



import dataclasses as _dc
import datetime as _dt
import enum as _enum
import functools as _ft
import time as _time
import uuid as _uuid
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    Iterator,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    no_type_check,
    runtime_checkable,
)

if TYPE_CHECKING:
    from ._aggregated_stream import AggregatedStream


@runtime_checkable
class MessageProtocol(Protocol):
    """
    Message protocol is a base class for all messages that are used in the system.
    """

    def get_message_id(self) -> _uuid.UUID:
        """
        Returns message ID
        """
        ...

    def get_message_time(self) -> _dt.datetime:
        """
        Returns message time
        """
        ...


E = TypeVar("E", bound=MessageProtocol)
M = TypeVar("M")


@_dc.dataclass(frozen=True)
class StoredMessage(Generic[E]):
    """
    Stored message is a message that is stored in the stream.

    Attributes:
        message_id: Message ID
        stream: Stream name
        version: Message version
        message: Message (`E` subtype of `MessageProtocol`)
        global_position: Global position
        added_at: DB timestamp when the message was added to the stream
    """

    message_id: _uuid.UUID
    stream: str
    version: int
    message: E
    global_position: int
    added_at: _dt.datetime


class AckOpProtocol(Protocol):
    def execute(self, **kwargs):
        ...


@_dc.dataclass(frozen=True)
class SubscriptionMessage(Generic[E]):
    """
    Subscription message is a message that is received from the subscription.

    Attributes:
        partition: Partition number
        position: Position in the partition
        stored_message: Stored message (`E` subtype of `MessageProtocol`)
    """

    partition: int
    position: int
    stored_message: StoredMessage[E]
    ack: AckOpProtocol


@_dc.dataclass()
class SubscriptionMessageBatch(Generic[E]):
    """

    Attributes:
    """

    partition: int
    first_position: int
    last_position: int
    lock_key: str
    messages: list[SubscriptionMessage[E]]
    ackd_position: int = -1

    @_ft.cached_property
    def message_ids(self) -> set[_uuid.UUID]:
        return {message.stored_message.message_id for message in self.messages}

    def ack(self, message: SubscriptionMessage[E]):
        """
        Acknowledges a message. This method is used to acknowledge messages in a batch.
        """
        if message.stored_message.message_id in self.message_ids:
            if self.ackd_position == -1:
                self.ackd_position = message.position
            else:
                assert (
                    self.ackd_position + 1 == message.position
                ), "You cannot leave gaps in the acknowledged messages"
                self.ackd_position = message.position
        else:
            raise ValueError("Message not found in the batch")


@_dc.dataclass(frozen=True)
class MessagePosition:
    """
    Message position is a position of the message in the stream.

    Attributes:
        stream: Stream name
        version: Message version
        global_position: Global position
    """

    stream: str
    version: int
    global_position: int


@_dc.dataclass
class StreamPartitionStatistic:
    partition_number: int
    next_message_id: _uuid.UUID
    next_message_position: int
    next_message_occurred_at: _dt.datetime
    max_position: int


@_dc.dataclass
class AggregatedStreamMessage:
    """
    Aggregated stream message is a message that can be read from an aggregated stream.

    Attributes:
        message_id: Message ID
        partition: Partition number
        position: Position in the partition
    """

    partition: int
    position: int
    message_id: _uuid.UUID


@_dc.dataclass
class SubscriptionState:
    """
    Subscription state is a state of the subscription.

    Attributes:
        positions: Mapping of partition number to the position in the partition
    """

    positions: Dict[int, int]


class MessageSerializer(Protocol, Generic[M]):
    """
    Message serializer is a protocol that is used to serialize and deserialize messages.

    The following must be true for any serializer:

    - `deserialize(serialize(message)) == message`
    - `type(deserialize(serialize(message))) is type(message)`
    - `serialize(deserialize(data)) == data`
    """

    def serialize(self, message: M) -> dict:
        """
        Serializes message to a dictionary. The dictionary must be JSON serializable.
        """
        raise NotImplementedError()

    def deserialize(self, message: dict) -> M:
        """
        Deserializes message from a dictionary.
        """
        raise NotImplementedError()


class MessagePartitioner(Protocol, Generic[E]):
    """
    Message partitioner is a protocol that is used to determine partition number for a message.
    """

    def get_partition(self, message: StoredMessage[E]) -> int:
        """
        Returns partition number for a message. The partition number must be a
        positive integer. The partition number must be deterministic for a given message.
        """
        raise NotImplementedError


class LockProvider(Protocol):
    """
    Lock provider is a protocol that is used to lock and unlock resources.
    """

    def lock(self, name: str) -> bool:
        """
        Locks resource with a given name. Returns `True` if the resource was locked.
        This method must not block!
        """
        raise NotImplementedError

    def unlock(self, name: str):
        """
        Unlocks resource with a given name.
        """
        raise NotImplementedError


class SubscriptionStateProvider(Protocol):
    """
    Subscription state provider is a protocol that is used to store and read subscription state.
    """

    def store(self, subscription_name: str, partition: int, position: int):
        """
        Stores subscription state for a given partition.
        """
        raise NotImplementedError

    def read(self, subscription_name: str) -> SubscriptionState:
        """
        Reads subscription state.

        Returns:
            Subscription state
        """
        raise NotImplementedError

    def initialize(self, subscription_name: str):
        """
        Marks subscription state as initialized.
        """
        raise NotImplementedError

    def initialized(self, subscription_name: str) -> bool:
        """
        Returns `True` if the subscription state was already initialized.
        """
        raise NotImplementedError

    def session(self, **kwargs) -> "SubscriptionStateProvider":
        """
        Returns a session for the subscription state provider.

        This can be used to run the state reads/updates in a transaction
        provided by the client.

        See [DbSubscriptionStateProvider][depeche_db.tools.DbSubscriptionStateProvider]
        for an example implementation.
        """
        raise NotImplementedError


class CallMiddleware(Generic[E]):
    """
    Call middleware is a protocol that is used to wrap a call to a handler.

    Typical implementation:

        class MyCallMiddleware(CallMiddleware):
            def __init__(self, some_dependency):
                self.some_dependency = some_dependency

            def call(self, handler, message):
                # or use a DI container here
                handler(message, some_dependency=self.some_dependency)
    """

    def call(
        self,
        handler: Callable,
        message: Union[SubscriptionMessage[E], StoredMessage[E], E],
    ):
        """
        Calls a handler with a given message.

        The type of the message depends on the type annotation of the handler function.
        See [MessageHandlerRegister][depeche_db.MessageHandlerRegister] for more details.

        Args:
            handler: Handler
            message: Message to be passed to the handler
        """
        raise NotImplementedError


class TimeBudget(Protocol):
    def over_budget(self) -> bool:
        """
        Returns `True` if the budget is over.
        """
        raise NotImplementedError


class FixedTimeBudget:
    def __init__(self, seconds: float):
        self.seconds = seconds
        self.start_time = _time.time()

    def over_budget(self) -> bool:
        """
        Returns `True` if the budget is over.
        """
        return _time.time() - self.start_time > self.seconds


class RunOnNotificationResult(_enum.Enum):
    WORK_REMAINING = "work_remaining"
    DONE_FOR_NOW = "done_for_now"


class RunOnNotification(Protocol):
    """
    Run on notification is a protocol that allows objects to be run when a
    notification is received on a channel. Objects that implement this protocol
    can be registered with a [Executor][depeche_db.Executor] object.

    Implemented by:
        - [SubscriptionRunner][depeche_db.SubscriptionRunner]
        - [StreamProjector][depeche_db.StreamProjector]
    """

    @property
    def notification_channel(self) -> str:
        """
        Returns notification channel name.
        """
        raise NotImplementedError

    def interested_in_notification(self, notification: dict) -> bool:
        """
        Signals whether the object is interested in a notification.
        If this method returns `False`, the object's `run` method will not be called
        because of this notification.
        """
        raise NotImplementedError

    def take_notification_hint(self, notification: dict):
        """
        This will be called when a notification is received. Make sure that
        the implementation of this method is fast, as it will be called for every
        notification received on the channel.
        Also make sure that this method is thread-safe, as it can be called
        from different threads. Especially, it should be thread-safe against
        the `run` method.
        """
        raise NotImplementedError

    def run(self, budget: TimeBudget) -> Optional[RunOnNotificationResult]:
        """
        Runs the object. This method needs to return when a chunk of work has been
        done.
        It needs to return within reasonable time after the given time budget is over.

        Returns:
            WORK_REMAINING if there is still work to be done
            DONE_FOR_NOW if there is no work to be done
            None will be interpreted as DONE_FOR_NOW (backwards compatibility)
        """
        raise NotImplementedError

    def stop(self):
        """
        If the object's `run` method has a loop, this method can be used to
        exit the loop earlier.
        Will be called in a separate thread.
        """
        raise NotImplementedError


class ErrorAction(_enum.Enum):
    """
    Error action is an action that is taken when an error occurs during message processing.

    Attributes:
        IGNORE: Ignore the error and continue processing.
        EXIT: Exit processing.
    """

    IGNORE = "ignore"
    EXIT = "exit"


class SubscriptionErrorHandler(Generic[E]):
    """
    Subscription error handler is a protocol that is used to handle errors that occur.
    """

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        """
        Handles an error that occurred during message processing.

        Args:
            error: Error
            message: Message that was being processed when the error occurred

        Returns:
            Action to be taken
        """
        raise NotImplementedError


class SubscriptionStartPoint:
    """
    Defines the start point of a subscription.
    """

    def init_state(
        self,
        subscription_name: str,
        stream: "AggregatedStream",
        state_provider: SubscriptionStateProvider,
    ):
        """
        Initializes subscription state (if not yet initialized).
        """
        raise NotImplementedError


@_dc.dataclass
class HandlerDescriptor(Generic[E]):
    handler: Callable
    pass_subscription_message: bool
    pass_stored_message: bool
    requires_middleware: bool

    @no_type_check
    def adapt_message_type(
        self, message: Union[SubscriptionMessage[E], StoredMessage[E], E]
    ) -> Union[SubscriptionMessage[E], StoredMessage[E], E]:
        if isinstance(message, SubscriptionMessage):
            if self.pass_subscription_message:
                return message
            elif self.pass_stored_message:
                return message.stored_message
            else:
                return message.stored_message.message
        elif isinstance(message, StoredMessage):
            if self.pass_subscription_message:
                raise ValueError(
                    "SubscriptionMessage was requested, but StoredMessage was provided"
                )
            elif self.pass_stored_message:
                return message.stored_message
            else:
                return message.stored_message.message
        else:
            if self.pass_subscription_message or self.pass_stored_message:
                raise ValueError(
                    "SubscriptionMessage or StoredMessage was requested, but plain message was provided"
                )
            else:
                return message


class MessageHandlerRegisterProtocol(Protocol, Generic[E]):
    """
    Message handler register protocol is used by runners to find handlers for messages.

    Implemented by:
        - [MessageHandlerRegister][depeche_db.MessageHandlerRegister]
        - [MessageHandler][depeche_db.MessageHandler]
    """

    def get_all_handlers(self) -> Iterator[HandlerDescriptor[E]]:
        """
        Returns all registered handlers.
        """
        raise NotImplementedError

    def get_handler(self, message_type: Type[E]) -> Optional[HandlerDescriptor[E]]:
        """
        Returns a handler for a given message type.
        """
        raise NotImplementedError



import datetime as _dt
import decimal as _decimal
import enum as _enum
import uuid as _uuid
from typing import Any, Callable, Optional, Type, TypeVar, no_type_check

import pydantic as _pydantic

from depeche_db._compat import get_union_members, issubclass_with_union

from .._interfaces import MessageSerializer

pydantic_v2 = _pydantic.VERSION.split(".")[0] == "2"
T = TypeVar("T", bound=_pydantic.BaseModel)


def _get_de_serializer(
    type_: Type[T],
) -> tuple[Callable[[dict], T], Callable[[T], dict]]:
    if pydantic_v2:

        def deserializer_v2(message: dict) -> T:
            return type_.model_validate(message)

        def serializer_v2(message: T) -> dict:
            return message.model_dump(mode="json")

        return deserializer_v2, serializer_v2

    def deserializer(message: dict) -> T:
        return type_.parse_obj(message)

    def serializer(message: T) -> dict:
        return make_jsonable(message.dict())  # type: ignore[no-any-return]

    return deserializer, serializer


class PydanticMessageSerializer(MessageSerializer[T]):
    # the real type would be: (self, message_type: Type[T])
    # but this is not supported by mypy (yet)
    def __init__(
        self, message_type: Any, aliases: Optional[dict[str, Type]] = None
    ) -> None:
        self.message_type: type[T] = message_type
        self.message_types: set[str] = set()
        self.serializers: dict[str, Callable[[Any], dict]] = {}
        self.deserializers: dict[str, Callable[[dict], Any]] = {}
        for member in get_union_members(message_type):
            if member.__name__ in self.message_types:
                raise ValueError(
                    f"Duplicate class name {member.__name__} in union {message_type}"
                )
            self.message_types.add(member.__name__)
            (
                self.deserializers[member.__name__],
                self.serializers[member.__name__],
            ) = _get_de_serializer(member)

        if aliases is not None:
            for alias, member in aliases.items():
                if alias in self.message_types:
                    raise ValueError(
                        f"Alias {alias} conflicts with existing class name in union {message_type}"
                    )
                if not issubclass_with_union(member, self.message_type):
                    raise ValueError(
                        f"Type {member} is not a subclass of {self.message_type}"
                    )
                self.message_types.add(alias)
                self.deserializers[alias], self.serializers[alias] = _get_de_serializer(
                    member
                )

    def serialize(self, message: T) -> dict:
        typename = message.__class__.__name__
        if typename not in self.message_types:
            raise TypeError(
                f"Unknown message type {typename} (expected one of {self.message_types})"
            )
        result: dict = self.serializers[typename](message)
        result["__typename__"] = typename
        return result

    @no_type_check
    def deserialize(self, message: dict) -> T:
        typename = message.pop("__typename__")
        if typename not in self.message_types:
            raise TypeError(
                f"Unknown message type {typename} (expected one of {self.message_types})"
            )
        return self.deserializers[typename](message)


def make_jsonable(obj):
    if isinstance(obj, list):
        return [make_jsonable(elem) for elem in obj]
    if isinstance(obj, dict):
        return {key: make_jsonable(value) for key, value in obj.items()}
    if isinstance(obj, _uuid.UUID):
        return str(obj)
    if isinstance(obj, _decimal.Decimal):
        return str(obj)
    if isinstance(obj, _dt.datetime):
        return obj.isoformat()
    if isinstance(obj, _dt.date):
        return obj.isoformat()
    if isinstance(obj, _enum.Enum):
        return obj.value
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj
    raise NotImplementedError(f"Cannot make object of type {type(obj)} jsonable")



import datetime as _dt
import enum as _enum
import logging as _logging
import random as _random
import time as _time
from typing import (
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)

from . import tools as _tools
from ._aggregated_stream import AggregatedStream
from ._compat import SAConnection
from ._interfaces import (
    CallMiddleware,
    ErrorAction,
    LockProvider,
    MessageHandlerRegisterProtocol,
    MessageProtocol,
    RunOnNotificationResult,
    StoredMessage,
    SubscriptionErrorHandler,
    SubscriptionMessage,
    SubscriptionMessageBatch,
    SubscriptionStartPoint,
    SubscriptionStateProvider,
    TimeBudget,
)

E = TypeVar("E", bound=MessageProtocol)

DEPECHE_LOGGER = _logging.getLogger("depeche_db")


class ExitSubscriptionErrorHandler(SubscriptionErrorHandler):
    """
    Exit the subscription on error
    """

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        return ErrorAction.EXIT


class LogAndIgnoreSubscriptionErrorHandler(SubscriptionErrorHandler):
    """
    Log the error and ignore the message
    """

    def __init__(self, subscription_name: str):
        self._logger = _logging.getLogger(
            f"depeche_db.subscription.{subscription_name}"
        )

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        self._logger.exception(
            f"Error while handling message {message.stored_message.message_id}:{message.stored_message.message.__class__.__name__}"
        )
        return ErrorAction.IGNORE


class AckStrategy(_enum.Enum):
    SINGLE = "single"
    BATCHED = "batched"


class NoAckOp:
    def execute(self, **kwargs):
        raise RuntimeError(
            "NoAckOp cannot be executed, choose AckStrategy.SINGLE if you want to use this operation"
        )


class AckRolledback(Exception):
    pass


class AckOp:
    def __init__(
        self,
        name: str,
        partition: int,
        position: int,
        state_provider: SubscriptionStateProvider,
    ):
        self.name = name
        self.partition = partition
        self.position = position
        self._state_provider = state_provider
        self._executed = False
        self._rolled_back = False

    def execute(self, **subscription_state_provider_kwargs):
        if self._executed:
            self._check()
            return

        if subscription_state_provider_kwargs:
            provider = self._state_provider.session(
                **subscription_state_provider_kwargs
            )
        else:
            provider = self._state_provider

        self._executed = True
        provider.store(
            subscription_name=self.name,
            partition=self.partition,
            position=self.position,
        )

    def _check(self):
        state = self._state_provider.read(self.name)
        if state.positions.get(self.partition, -1) != self.position:
            raise AckRolledback()

    def rollback(self):
        self._rolled_back = True

    @property
    def rolled_back(self):
        return self._rolled_back


class Subscription(Generic[E]):
    _state_provider: SubscriptionStateProvider
    runner: "Union[SubscriptionRunner[E], BatchedAckSubscriptionRunner[E]]"

    def __init__(
        self,
        name: str,
        stream: AggregatedStream[E],
        message_handler: "SubscriptionMessageHandler[E]",
        batch_size: Optional[int] = None,
        state_provider: Optional[SubscriptionStateProvider] = None,
        lock_provider: Optional[LockProvider] = None,
        start_point: Optional[SubscriptionStartPoint] = None,
        ack_strategy: AckStrategy = AckStrategy.SINGLE,
    ):
        """
        A subscription is a way to read messages from an aggregated stream.

        Read more about the subscription in the [concepts section](../concepts/subscriptions.md).

        Args:
            name: Name of the subscription, needs to be a valid python identifier
            stream: Stream to read from
            message_handler: Handler for the messages
            batch_size: Number of messages to read at once, defaults to 10, read more [here][depeche_db.SubscriptionRunner]
            state_provider: Provider for the subscription state, defaults to a PostgreSQL provider
            lock_provider: Provider for the locks, defaults to a PostgreSQL provider
            start_point: The start point for the subscription, defaults to beginning of the stream
            ack_strategy: The strategy to use for acknowledging messages, defaults to AckStrategy.SINGLE.
        """
        assert name.isidentifier(), "Group name must be a valid identifier"
        self.name = name
        self._stream = stream
        self._lock_provider = lock_provider or _tools.DbLockProvider(
            name, self._stream._store.engine
        )
        self._state_provider = state_provider or _tools.DbSubscriptionStateProvider(
            name, self._stream._store.engine
        )
        self._start_point = start_point
        if ack_strategy == AckStrategy.BATCHED:
            self.runner = BatchedAckSubscriptionRunner(
                subscription=self,
                message_handler=message_handler,
                batch_size=batch_size,
            )
        elif ack_strategy == AckStrategy.SINGLE:
            self.runner = SubscriptionRunner(
                subscription=self,
                message_handler=message_handler,
                batch_size=batch_size,
            )
        else:
            raise NotImplementedError(f"Ack strategy {ack_strategy} is not implemented")

        self._max_aggregated_stream_positions_cache: Dict[int, int] = {}

    def _init_state(self):
        if not self._state_provider.initialized(self.name):
            lock_key = f"subscription-{self.name}-init"
            if not self._lock_provider.lock(lock_key):
                # another instance is already initializing the state
                # wait until it is initialized
                while not self._state_provider.initialized(self.name):
                    _time.sleep(0.05)
                return
            try:
                if self._start_point is not None:
                    self._start_point.init_state(
                        subscription_name=self.name,
                        stream=self._stream,
                        state_provider=self._state_provider,
                    )
                self._state_provider.initialize(self.name)
            finally:
                self._lock_provider.unlock(lock_key)

    def _update_max_aggregated_stream_positions_cache(
        self, partition_number: int, max_position: int
    ) -> None:
        """
        Updates the cache of max aggregated stream positions for a specific partition.
        This is used to keep the cache up-to-date when new messages are added to the stream.
        """
        current_value = self._max_aggregated_stream_positions_cache.get(
            partition_number, -1
        )
        self._max_aggregated_stream_positions_cache[partition_number] = max(
            max_position, current_value
        )

    def _get_next_partitions(self, conn: SAConnection) -> List[int]:
        state = self._state_provider.read(self.name)

        def _refresh_max_aggregated_stream_positions_cache():
            self._max_aggregated_stream_positions_cache = self._stream._get_max_aggregated_stream_positions(
                conn=conn,
                # TODO how to make sure we have all partitions in the state?
                # min_position=min(
                #    (position - 1000 for position in state.positions), default=0
                # ),
            )

        def _calculate_unprocessed_message_counts():
            if not self._max_aggregated_stream_positions_cache:
                return {}

            unprocessed_message_counts = []
            for (
                partition_number,
                max_position,
            ) in self._max_aggregated_stream_positions_cache.copy().items():
                current_position = state.positions.get(partition_number, -1)
                if current_position < max_position:
                    # there are still messages to read in this partition
                    unprocessed_message_counts.append(
                        (max_position - current_position, partition_number)
                    )
            return unprocessed_message_counts

        unprocessed_message_counts = _calculate_unprocessed_message_counts()
        if not unprocessed_message_counts:
            _refresh_max_aggregated_stream_positions_cache()
            unprocessed_message_counts = _calculate_unprocessed_message_counts()

        # Take the top 20 partitions with the most messages to read
        result = [
            partition_number
            for _, partition_number in sorted(
                unprocessed_message_counts,
                reverse=True,
            )
        ]
        result = result[:20]
        # Shuffle the partitions to avoid reading them in the same order on multiple instances
        _random.shuffle(result)
        return result

    def get_next_message_batch(
        self, count: int
    ) -> Optional[SubscriptionMessageBatch[E]]:
        if not self._state_provider.initialized(self.name):
            self._init_state()
        assert self._state_provider.initialized(self.name)

        with self._stream._store.engine.connect() as conn:
            for partition_number in self._get_next_partitions(conn=conn):
                lock_key = f"subscription-{self.name}-{partition_number}"
                if not self._lock_provider.lock(lock_key):
                    continue

                # now we have the lock, we need to get the current state of
                # the partition to determine where to start reading
                state = self._state_provider.read(self.name)
                current_position = state.positions.get(partition_number, -1)
                next_message_position = current_position + 1

                message_pointers = list(
                    self._stream.read_slice(
                        partition=partition_number,
                        start=next_message_position,
                        count=count,
                        conn=conn,
                    )
                )
                if not message_pointers:
                    # No messages -> try the next partition
                    self._lock_provider.unlock(lock_key)
                    continue

                with self._stream._store.reader(conn=conn) as reader:
                    stored_messages = {
                        message.message_id: message
                        for message in reader.get_messages_by_ids(
                            [pointer.message_id for pointer in message_pointers]
                        )
                    }
                messages = [
                    SubscriptionMessage(
                        partition=pointer.partition,
                        position=pointer.position,
                        stored_message=stored_messages[pointer.message_id],
                        ack=NoAckOp(),
                    )
                    for pointer in message_pointers
                ]
                return SubscriptionMessageBatch(
                    partition=partition_number,
                    first_position=min(msg.position for msg in messages),
                    last_position=max(msg.position for msg in messages),
                    lock_key=lock_key,
                    messages=messages,
                )
            return None

    def ack_message_batch(
        self, message_batch: SubscriptionMessageBatch[E], success: bool
    ) -> None:
        if success:
            self._state_provider.store(
                subscription_name=self.name,
                partition=message_batch.partition,
                position=message_batch.ackd_position,
            )
        self.unlock_message_batch(message_batch)

    def unlock_message_batch(self, message_batch: SubscriptionMessageBatch[E]) -> None:
        self._lock_provider.unlock(message_batch.lock_key)

    def get_next_messages(self, count: int) -> Iterator[SubscriptionMessage[E]]:
        batch = None
        try:
            batch = self.get_next_message_batch(count=count)
            if batch:
                for message in batch.messages:
                    ack = AckOp(
                        name=self.name,
                        partition=message.partition,
                        position=message.position,
                        state_provider=self._state_provider,
                    )

                    yield SubscriptionMessage(
                        partition=message.partition,
                        position=message.position,
                        stored_message=message.stored_message,
                        ack=ack,
                    )
                    if ack.rolled_back:
                        # the message was not ack'd or the acknolwedgement was rolled back
                        break
                    try:
                        ack.execute()
                    except AckRolledback:
                        # the message was not ack'd or the acknolwedgement was rolled back
                        break
        finally:
            if batch:
                self.unlock_message_batch(batch)


class SubscriptionMessageHandler(Generic[E]):
    def __init__(
        self,
        handler_register: MessageHandlerRegisterProtocol[E],
        error_handler: Optional[SubscriptionErrorHandler] = None,
        call_middleware: Optional[CallMiddleware] = None,
    ):
        """
        Handles messages

        Args:
            handler_register: The handler register to use
            error_handler: A handler for errors raised by the handlers, defaults to handler that will exit the subscription
            call_middleware: The middleware to call before calling the handler
        """
        self._register = handler_register
        self._error_handler = error_handler or ExitSubscriptionErrorHandler()
        self._call_middleware = call_middleware

        if not self._call_middleware and any(
            handler.requires_middleware for handler in self._register.get_all_handlers()
        ):
            raise ValueError(
                "If handler has more than one parameter, a call_middleware must be provided"
            )

    def handle(self, message: SubscriptionMessage):
        handler = self._register.get_handler(type(message.stored_message.message))
        if handler:
            try:
                self._exec(handler.handler, handler.adapt_message_type(message))
            except Exception as error:
                error_handling_result = self._error_handler.handle_error(
                    error=error, message=message
                )
                if error_handling_result == ErrorAction.EXIT:
                    raise

    def _exec(
        self,
        handler: Callable[..., None],
        message: Union[SubscriptionMessage[E], StoredMessage[E], E],
    ) -> None:
        if self._call_middleware:
            self._call_middleware.call(handler, message)
        else:
            handler(message)


class SubscriptionRunner(Generic[E]):
    def __init__(
        self,
        subscription: Subscription[E],
        message_handler: SubscriptionMessageHandler,
        batch_size: Optional[int] = None,
    ):
        """
        Handles messages from a subscription using a handler

        The `batch_size` argument controls how many messages to handle in each
        batch. If not provided, the default is 10. A larger batch size will
        result less round trips to the database, but will also make it more
        likely that messages from _different partitions_ will be processed out of
        the order defined by their `global_position` on the message store.

        A batch size of 1 will ensure that messages are processed in order
        regarding to their `global_position`.
        Messages in the same partition will always be processed in order.

        Implements: [RunOnNotification][depeche_db.RunOnNotification]

        Args:
            subscription: The subscription to handle
            message_handler: The handler to use
            batch_size: The number of messages to handle in each batch, defaults to 10
        """
        self._subscription = subscription
        self._batch_size = batch_size or 10
        self._keep_running = True
        self._handler = message_handler

    def interested_in_notification(self, notification: dict) -> bool:
        return True

    def take_notification_hint(self, notification: dict):
        partition_number = notification.get("partition")
        position = notification.get("position")
        if partition_number is not None and position is not None:
            self._subscription._update_max_aggregated_stream_positions_cache(
                partition_number, position
            )

    @property
    def notification_channel(self) -> str:
        return self._subscription._stream.notification_channel

    def run(self, budget: Optional[TimeBudget] = None) -> RunOnNotificationResult:
        return self.run_once(budget=budget)

    def stop(self):
        self._keep_running = False

    def run_once(self, budget: Optional[TimeBudget] = None) -> RunOnNotificationResult:
        while self._keep_running:
            n = 0
            for message in self._subscription.get_next_messages(count=self._batch_size):
                n += 1
                self.handle(message)
            if n == 0:
                break
            if budget and budget.over_budget():
                return RunOnNotificationResult.WORK_REMAINING
        return RunOnNotificationResult.DONE_FOR_NOW

    def handle(self, message: SubscriptionMessage):
        self._handler.handle(message)


class BatchedAckSubscriptionRunner(SubscriptionRunner[E]):
    def run_once(self, budget: Optional[TimeBudget] = None) -> RunOnNotificationResult:
        while self._keep_running:
            message_batch = self._subscription.get_next_message_batch(
                count=self._batch_size
            )
            if message_batch is None:
                break
            try:
                for message in message_batch.messages:
                    self.handle(message)
                    message_batch.ack(message)
            finally:
                self._subscription.ack_message_batch(
                    message_batch=message_batch, success=True
                )
            if budget and budget.over_budget():
                return RunOnNotificationResult.WORK_REMAINING
        return RunOnNotificationResult.DONE_FOR_NOW


class StartAtNextMessage(SubscriptionStartPoint):
    """
    Starts consuming messages from the next message in the stream.
    """

    def init_state(
        self,
        subscription_name: str,
        stream: "AggregatedStream",
        state_provider: SubscriptionStateProvider,
    ):
        for partition_statistic in stream.get_partition_statistics():
            state_provider.store(
                subscription_name=subscription_name,
                partition=partition_statistic.partition_number,
                position=partition_statistic.max_position,
            )


class StartAtPointInTime(SubscriptionStartPoint):
    def __init__(self, point_in_time: _dt.datetime):
        """
        Starts consuming messages from a point in time.

        Args:
            point_in_time: The point in time to start consuming messages from. The point in time must be timezone aware.
        """
        if not point_in_time.tzinfo:
            raise ValueError("Point in time must be timezone aware")
        self._point_in_time = point_in_time

    def init_state(
        self,
        subscription_name: str,
        stream: "AggregatedStream",
        state_provider: SubscriptionStateProvider,
    ):
        for partition, position in stream.time_to_positions(
            self._point_in_time
        ).items():
            if position > 0:
                state_provider.store(
                    subscription_name=subscription_name,
                    partition=partition,
                    position=position - 1,
                )



import dataclasses as _dc
import json
import logging
import queue
import select
import threading
import time
from typing import Iterator, Sequence

from depeche_db._compat import PSYCOPG_VERSION
from depeche_db._compat import psycopg as _psycopg

logger = logging.getLogger(__name__)


@_dc.dataclass
class PgNotification:
    channel: str
    payload: dict


class PgNotificationListener:
    def __init__(
        self,
        dsn: str,
        channels: Sequence[str],
        select_timeout: float = 3.0,
        ignore_payload: bool = False,
    ):
        self.dsn = dsn
        self.channels = channels
        self._keep_running = True
        self._thread = threading.Thread(target=self._loop)
        self._queue = queue.Queue[PgNotification]()
        self._ignore_payload = ignore_payload
        self._select_timeout = select_timeout
        self._queue_timeout = select_timeout / 2

    def messages(self, timeout: float = 0) -> Iterator[PgNotification]:
        last_message_at = time.time()
        while self._keep_running:
            try:
                yield self._queue.get(block=True, timeout=self._queue_timeout)
                last_message_at = time.time()
            except queue.Empty:
                if timeout > 0:
                    if time.time() - last_message_at > timeout:
                        break
                pass

    def start(self):
        self._thread.start()

    def stop(self):
        self._keep_running = False
        self._thread.join()

    def _parse_dsn(self, dsn: str) -> tuple[str, str]:
        import urllib.parse as _urlparse

        parts = list(_urlparse.urlparse(dsn))
        expected_psycopg_version = "2"
        if parts[0] == "postgresql+psycopg":
            parts[0] = "postgresql"
            expected_psycopg_version = "3"
        if parts[0] == "postgresql+psycopg2":
            parts[0] = "postgresql"
        dsn = _urlparse.urlunparse(parts)

        return dsn, expected_psycopg_version

    def _loop(self):
        dsn, expected_psycopg_version = self._parse_dsn(self.dsn)
        assert expected_psycopg_version == PSYCOPG_VERSION, "Invalid psycopg version"
        conn = _psycopg.connect(dsn)

        try:
            curs = conn.cursor()
            for channel in self.channels:
                curs.execute(f"LISTEN {channel};")
            curs.execute("commit;")

            while self._keep_running:
                if select.select([conn], [], [], self._select_timeout) == ([], [], []):
                    pass
                else:
                    if PSYCOPG_VERSION == "3":
                        self._psycopg3_loop(conn)
                    else:
                        self._psycopg2_loop(conn)
        finally:
            conn.close()

    def _psycopg3_loop(self, conn):
        for notification in conn.notifies(timeout=0.2):
            self._process_notification(notification)

    def _psycopg2_loop(self, conn):
        conn.poll()
        while conn.notifies:
            self._process_notification(conn.notifies.pop(0))

    def _process_notification(self, notification):
        try:
            if self._ignore_payload:
                payload = {}
            else:
                payload = json.loads(notification.payload)
            self._queue.put(PgNotification(notification.channel, payload))
        except Exception:
            logger.exception(f"Error processing notification payload: {notification}")



import collections as _collections
import dataclasses as _dc
import queue as _queue
import signal as _signal
import threading as _threading
import time as _time
import uuid as _uuid
from typing import Dict, List, Optional

from .._executor import UniqueQueue
from .._interfaces import FixedTimeBudget, RunOnNotification, RunOnNotificationResult
from ..tools import PgNotificationListener


@_dc.dataclass
class HandlerRegistration:
    handler: RunOnNotification
    concurrency: int
    id: _uuid.UUID = _dc.field(default_factory=_uuid.uuid4)


class ThreadedExecutor:
    """
    Executor is a class that runs handlers on notifications.

    This version of the executor runs handlers in separate threads, allowing
    for concurrent processing of notifications.

    This is an experimental feature and may
      a) not be suitable for all use cases and
      b) change without notice in future versions.

    Typical usage:

        executor = ThreadedExecutor(db_dsn="postgresql://localhost:5432/mydb")
        executor.register(MyRunOnNotification())
        executor.run()   # this will stop on SIGTERM or SIGINT

    Args:
        db_dsn: DSN for the PostgreSQL database
        stimulation_interval: Seconds; Every interval, all the handlers will run once. Zero & negative values disable stimulation.
        disable_signals: Disable catching SIGINT/SIGTERM, useful if you want to run the executor in a sub-thread
    """

    listener: Optional[PgNotificationListener] = None

    def __init__(
        self,
        db_dsn: str,
        stimulation_interval: float = 0.5,
        disable_signals: bool = False,
    ):
        self._db_dsn = db_dsn
        self.channel_register: Dict[
            str, List[HandlerRegistration]
        ] = _collections.defaultdict(list)
        self.stimulation_interval = stimulation_interval
        self.keep_running = True
        self.handler_queues: Dict[_uuid.UUID, UniqueQueue] = {}
        self.handler_threads: List[_threading.Thread] = []
        self.stimulator_thread = _threading.Thread(target=self._stimulate, daemon=True)
        self.listener = None
        if not disable_signals:
            _signal.signal(_signal.SIGINT, lambda *_: self._stop())
            _signal.signal(_signal.SIGTERM, lambda *_: self._stop())

    def register(self, handler: RunOnNotification):
        """
        Registers a handler to be run on notifications.

        Args:
            handler: Handler to register
        """
        concurrency = 1  # TODO v2: allow setting concurrency
        self.channel_register[handler.notification_channel].append(
            HandlerRegistration(handler=handler, concurrency=concurrency)
        )
        return handler

    def _stop(self):
        self.keep_running = False
        if self.listener is not None:
            self.listener.stop()

        for registrations in self.channel_register.values():
            for registration in registrations:
                registration.handler.stop()

    def run(self):
        """Runs the executor."""
        self.listener = PgNotificationListener(
            dsn=self._db_dsn,
            channels=list(self.channel_register),
            ignore_payload=False,
        )

        for registrations in self.channel_register.values():
            for registration in registrations:
                self.handler_queues[registration.id] = UniqueQueue()
                for _ in range(registration.concurrency):
                    self.handler_threads.append(
                        _threading.Thread(
                            target=self._run_handler(registration),
                            daemon=True,
                        )
                    )
        for thread in self.handler_threads:
            thread.start()
        self.stimulator_thread.start()
        self.listener.start()

        for notification in self.listener.messages():
            for registration in self.channel_register[notification.channel]:
                if registration.handler.interested_in_notification(
                    notification.payload
                ):
                    registration.handler.take_notification_hint(notification.payload)
                    self.handler_queues[registration.id].put(registration.id)

        for thread in self.handler_threads:
            thread.join()
        self.stimulator_thread.join()

    def _run_handler(self, registration: HandlerRegistration):
        def run_handler():
            while self.keep_running:
                try:
                    self.handler_queues[registration.id].get(timeout=0.5)
                    result = registration.handler.run(budget=FixedTimeBudget(seconds=3))
                    result = result or RunOnNotificationResult.DONE_FOR_NOW
                    if result == RunOnNotificationResult.WORK_REMAINING:
                        # Re-queue the handler if it has work remaining
                        self.handler_queues[registration.id].put(registration.id)
                except _queue.Empty:
                    pass

        return run_handler

    def _stimulate(self):
        if self.stimulation_interval <= 0:
            return
        while self.keep_running:
            for handlers in self.channel_register.values():
                for handler in handlers:
                    self.handler_queues[handler.id].put(handler.id)

            # Wait for the stimulation interval to pass
            started_at = _time.time()
            while (
                self.keep_running
                and _time.time() - started_at < self.stimulation_interval
            ):
                _time.sleep(0.1)



from .._interfaces import LockProvider, SubscriptionState, SubscriptionStateProvider


class InMemorySubscriptionState(SubscriptionStateProvider):
    def __init__(self):
        self._state = {}

    def store(self, subscription_name: str, partition: int, position: int):
        if subscription_name not in self._state:
            self._state[subscription_name] = {}

        self._state[subscription_name][partition] = position

    def read(self, subscription_name: str) -> SubscriptionState:
        if subscription_name not in self._state:
            return SubscriptionState({})
        return SubscriptionState(self._state[subscription_name])

    def initialize(self, subscription_name: str):
        if subscription_name not in self._state:
            self._state[subscription_name] = {}

    def initialized(self, subscription_name: str) -> bool:
        return subscription_name in self._state

    def session(self, **kwargs) -> SubscriptionStateProvider:
        return self


class NullLockProvider(LockProvider):
    def __init__(self):
        self._locks = {}

    def lock(self, name: str) -> bool:
        return True

    def unlock(self, name: str):
        pass



from typing import Dict

import pals as _pals
import sqlalchemy as _sa


class DbLockProvider:
    def __init__(self, name: str, engine: _sa.engine.Engine):
        assert name.isidentifier(), "Name must be a valid identifier"
        self._engine = engine
        self.locker = _pals.Locker(
            app_name=name,
            create_engine_callable=lambda: self._engine,
            blocking_default=False,
        )
        self._locks: Dict[str, _pals.Lock] = {}  # type: ignore

    def lock(self, name: str) -> bool:
        # assert name not in self._locks, "Lock already acquired"
        if name in self._locks:
            return False
        lock = self._locks[name] = self.locker.lock(name, blocking=False)
        result = lock.acquire()
        if not result:
            if lock.conn:
                lock.conn.close()
            del self._locks[name]
        return result  # type: ignore

    def unlock(self, name: str):
        lock = self._locks.pop(name)
        lock.release()

    def finalize(self):
        try:
            self.locker.engine.pool.dispose()
        except Exception:
            pass

    def __del__(self):
        self.finalize()



from typing import TypeVar

from .._interfaces import MessagePartitioner, MessageProtocol, StoredMessage

E = TypeVar("E", bound=MessageProtocol)


class StreamNamePartitioner(MessagePartitioner[E]):
    def get_partition(self, message: StoredMessage[E]) -> int:
        import hashlib

        return int.from_bytes(
            hashlib.sha256(message.stream.encode()).digest()[:2], "little"
        )



import abc as _abc
from typing import Callable, Generic, TypeVar

from depeche_db import (
    MessagePosition,
    MessageProtocol,
    MessageStore,
    MessageStoreReader,
)

from .aggregate_root import EventSourcedAggregateRoot

E = TypeVar("E", bound=MessageProtocol)
OBJ = TypeVar("OBJ", bound=EventSourcedAggregateRoot)
ID = TypeVar("ID")


class Repo(_abc.ABC, Generic[OBJ, ID]):
    """
    A repository is a collection of objects that can be queried and
    persisted.

    This is an abstract base class that defines the interface that
    all repositories implement.
    """

    @_abc.abstractmethod
    def add(self, entity: OBJ) -> MessagePosition:
        """
        Add a new entity to the repository.
        """
        raise NotImplementedError

    @_abc.abstractmethod
    def save(self, entity: OBJ, expected_version: int) -> MessagePosition:
        """
        Save an existing entity to the repository.
        """
        raise NotImplementedError

    @_abc.abstractmethod
    def get(self, id: ID) -> OBJ:
        """
        Get an entity from the repository by its ID.
        """
        raise NotImplementedError


class EventStoreRepo(Generic[E, OBJ, ID], Repo[OBJ, ID]):
    def __init__(
        self,
        event_store: MessageStore[E],
        constructor: Callable[[], OBJ],
        stream_prefix: str,
    ):
        self._event_store = event_store
        self._constructor = constructor
        self._stream_prefix = stream_prefix

    def add(self, obj: OBJ) -> MessagePosition:
        return self.save(obj, expected_version=0)

    def save(self, obj: OBJ, expected_version: int) -> MessagePosition:
        return self._event_store.synchronize(
            stream=f"{self._stream_prefix}-{obj.get_id()}",
            messages=obj.events,
            expected_version=expected_version,
        )

    def get(self, id: ID) -> OBJ:
        with self._event_store.reader() as reader:
            return ReadRepository[E, OBJ, ID](
                event_store_reader=reader,
                constructor=self._constructor,
                stream_prefix=self._stream_prefix,
            ).get(id)


class NotFound(Exception):
    pass


class ReadRepository(Generic[E, OBJ, ID]):
    def __init__(
        self,
        event_store_reader: MessageStoreReader[E],
        constructor: Callable[[], OBJ],
        stream_prefix: str,
    ):
        self._event_store_reader = event_store_reader
        self._constructor = constructor
        self._stream_prefix = stream_prefix

    def get(self, id: ID) -> OBJ:
        obj = self._constructor()
        for event in self._event_store_reader.read(f"{self._stream_prefix}-{id}"):
            obj.apply(event.message)
        if obj.get_id() is None:
            raise NotFound(id)
        return obj



import os
import sys
import typing

if sys.version_info < (3, 9):
    raise RuntimeError("DepecheDB requires Python 3.9 or newer")
elif sys.version_info < (3, 10):
    UNION_TYPES = (typing.Union,)

    def issubclass_with_union(cls, superclass):
        if typing.get_origin(superclass) in UNION_TYPES:
            if cls == superclass:
                return True
            return any(
                issubclass_with_union(cls, member)
                for member in typing.get_args(superclass)
            )

        if typing.get_origin(cls) in UNION_TYPES:
            return all(
                issubclass_with_union(member, superclass)
                for member in typing.get_args(cls)
            )

        return issubclass(cls, superclass)

else:
    import types

    UNION_TYPES = (typing.Union, types.UnionType)

    def issubclass_with_union(cls, class_or_tuple):
        if typing.get_origin(cls) in UNION_TYPES:
            return all(
                issubclass_with_union(member, class_or_tuple)
                for member in typing.get_args(cls)
            )
        return issubclass(cls, class_or_tuple)


def get_union_members(union_or_type) -> typing.Generator[typing.Type, None, None]:
    if typing.get_origin(union_or_type) in UNION_TYPES:
        for member in typing.get_args(union_or_type):
            yield from get_union_members(member)
    else:
        yield union_or_type


SA_VERSION = "2.x"
try:
    from sqlalchemy import Connection as SAConnection  # noqa
    from sqlalchemy import ColumnElement as SAColumnElement  # noqa
    from sqlalchemy import Row as SARow  # noqa
except ImportError:
    SA_VERSION = "1.4.x"
    from sqlalchemy.engine import Row as SARow  # noqa
    from sqlalchemy.engine import Connection as SAConnection  # noqa
    from sqlalchemy.sql.expression import ColumnElement as SAColumnElement  # noqa

PSYCOPG2_AVAILABLE = False
PSYCOPG3_AVAILABLE = False
PSYCOPG_VERSION: typing.Union[str, None] = None
try:
    if os.environ.get("DEPECHE_DB_FORCE_PSYCOPG3", "0") != "1":
        import psycopg2 as psycopg  # noqa
        from psycopg2.errors import LockNotAvailable as PsycoPgLockNotAvailable  # noqa
        from psycopg2.errors import RaiseException as PsycoPgRaiseException  # noqa
        from psycopg2.extras import Json as PsycoPgJson  # noqa

        PSYCOPG_VERSION = "2"
        PSYCOPG2_AVAILABLE = True
except ImportError:
    pass

if PSYCOPG_VERSION is None:
    try:
        import psycopg  # noqa
        from psycopg.errors import LockNotAvailable as PsycoPgLockNotAvailable  # noqa
        from psycopg.errors import RaiseException as PsycoPgRaiseException  # noqa
        from psycopg.types.json import Json as PsycoPgJson  # noqa

        PSYCOPG_VERSION = "3"
        PSYCOPG3_AVAILABLE = True
    except ImportError:
        raise RuntimeError(
            "DepecheDB requires psycopg2 or psycopg3; please install it: e.g. pip install psycopg2-binary"
        )
else:
    try:
        import psycopg as _psycopg3  # noqa

        PSYCOPG3_AVAILABLE = True
    except ImportError:
        pass



import argparse as _argparse
from typing import Tuple

from ._aggregated_stream import AggregatedStream


def main():
    """
    Commands:
        - generate-migration-script: Generate migration scripts for aggregated streams.
    """
    parser = _argparse.ArgumentParser(description="DepecheDB CLI tools")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Add generate-migration-script subcommand
    migration_parser = subparsers.add_parser(
        "generate-migration-script",
        help="Generate migration scripts for aggregated streams",
    )
    migration_parser.add_argument("prev_version", help="Previous version (e.g., '0.8')")
    migration_parser.add_argument(
        "target_version", help="Target version (e.g., '0.11')"
    )
    migration_parser.add_argument(
        "--message-store", required=True, help="Name of the message store"
    )
    migration_parser.add_argument(
        "--aggregated-stream",
        action="append",
        required=True,
        dest="aggregated_streams",
        help="Name of an aggregated stream (can be specified multiple times)",
    )

    args = parser.parse_args()

    if args.command == "generate-migration-script":
        migration_generators = AggregatedStream.migration_script_generators()
        prev_version = parse_version(args.prev_version)
        target_version = parse_version(args.target_version)

        for version, generators in migration_generators.items():
            if version < prev_version:
                continue
            if version > target_version:
                continue

            if generators:
                print("-- Migration for version:", ".".join(map(str, version)))

            for stream_name in args.aggregated_streams:
                print(f"-- Migration for aggregated stream: {stream_name}")
                for generator in generators:
                    migration_ddl = generator(stream_name, args.message_store)
                    print(migration_ddl)
                print()

    else:
        parser.print_help()


def parse_version(version_str: str) -> Tuple[int, int]:
    """
    Parse a version string into a tuple of integers.
    """
    major, minor, *_ = [int(part) for part in version_str.split(".")]
    return (major, minor)


if __name__ == "__main__":
    main()



import contextlib as _contextlib
from typing import Iterator, Set

import sqlalchemy as _sa

from .._compat import SAConnection
from .._interfaces import SubscriptionState, SubscriptionStateProvider


class DbSubscriptionStateProvider(SubscriptionStateProvider):
    SPECIAL_PARTITION_FOR_INIT_STATE = -1

    def __init__(self, name: str, engine: _sa.engine.Engine):
        assert name.isidentifier(), "Name must be a valid identifier"
        self.name = name
        self._engine = engine

        self.metadata = _sa.MetaData()
        self.state_table = _sa.Table(
            f"depeche_subscriptions_{name}",
            self.metadata,
            _sa.Column("subscription_name", _sa.String, primary_key=True),
            _sa.Column("partition", _sa.Integer, primary_key=True),
            _sa.Column("position", _sa.Integer, nullable=False),
        )
        self.metadata.create_all(self._engine)
        self._initialized_subscriptions: Set[str] = set()

    def initialize(self, subscription_name: str):
        """
        Marks subscription state as initialized.
        """
        self.store(
            subscription_name=subscription_name,
            partition=self.SPECIAL_PARTITION_FOR_INIT_STATE,
            position=-1,
        )

    def initialized(self, subscription_name: str) -> bool:
        """
        Returns `True` if the subscription state was already initialized.
        """
        # Caches the result if True. It cannot go to False again!
        if subscription_name in self._initialized_subscriptions:
            return True

        with self._engine.connect() as conn:
            result: bool = conn.execute(
                _sa.select(
                    self.state_table.c.partition.isnot(None),
                ).where(
                    _sa.and_(
                        self.state_table.c.subscription_name == subscription_name,
                        self.state_table.c.partition
                        == self.SPECIAL_PARTITION_FOR_INIT_STATE,
                    )
                )
            ).scalar()

        if result:
            self._initialized_subscriptions.add(subscription_name)
        return result

    def store(self, subscription_name: str, partition: int, position: int):
        with self._session() as session:
            session.store(
                subscription_name=subscription_name,
                partition=partition,
                position=position,
            )

    def read(self, subscription_name: str) -> SubscriptionState:
        with self._session() as session:
            return session.read(subscription_name=subscription_name)

    @_contextlib.contextmanager
    def _session(self) -> Iterator["_Session"]:
        with self._engine.connect() as conn:
            yield _Session(conn=conn, parent=self)
            conn.commit()

    def session(self, conn: _sa.engine.Connection = None, **kwargs):
        assert conn, "Connection is required"
        assert not kwargs, "More arguments are not supported"
        return _Session(conn=conn, parent=self)

    @classmethod
    def get_migration_ddl(cls, name: str):
        """
        DDL Script to migrate from <=0.8.0
        """
        tablename = (f"depeche_subscriptions_{name}",)
        return f"""
            ALTER TABLE "{name}_subscription_state"
                 RENAME TO {tablename};
            """

    @classmethod
    def migrate_db_objects(cls, name: str, conn: SAConnection):
        """
        Migrate from <=0.8.0
        """
        conn.execute(cls.get_migration_ddl(name=name))


class _Session:
    def __init__(
        self, conn: _sa.engine.Connection, parent: DbSubscriptionStateProvider
    ):
        self._conn = conn
        self._parent = parent

    def initialize(self, subscription_name: str):
        self._parent.initialize(subscription_name=subscription_name)

    def initialized(self, subscription_name: str) -> bool:
        return self._parent.initialized(subscription_name=subscription_name)

    def store(self, subscription_name: str, partition: int, position: int):
        from sqlalchemy.dialects.postgresql import insert

        state_table = self._parent.state_table
        self._conn.execute(
            insert(state_table)
            .values(
                subscription_name=subscription_name,
                partition=partition,
                position=position,
            )
            .on_conflict_do_update(
                index_elements=[
                    state_table.c.subscription_name,
                    state_table.c.partition,
                ],
                set_={
                    state_table.c.position: position,
                },
            )
        )

    def read(self, subscription_name: str) -> SubscriptionState:
        state_table = self._parent.state_table
        return SubscriptionState(
            {
                row.partition: row.position
                for row in self._conn.execute(
                    _sa.select(
                        state_table.c.partition,
                        state_table.c.position,
                    ).where(
                        _sa.and_(
                            state_table.c.subscription_name == subscription_name,
                            state_table.c.partition
                            != self._parent.SPECIAL_PARTITION_FOR_INIT_STATE,
                        )
                    )
                )
            }
        )



"""
The storage implementation in PL/SQL is heavily inspired by
https://github.com/message-db/message-db
"""

import datetime as _dt
import uuid as _uuid
from typing import Any, Iterator, Optional, Sequence, Tuple

import sqlalchemy as _sa
from sqlalchemy.dialects.postgresql import JSONB as _PostgresJsonb
from sqlalchemy_utils import UUIDType as _UUIDType

from ._compat import PsycoPgJson, SAConnection
from ._exceptions import OptimisticConcurrencyError
from ._interfaces import MessagePosition


class Storage:
    name: str

    def __init__(self, name: str, engine: _sa.engine.Engine):
        assert name.isidentifier(), "name must be a valid identifier"
        self.name = name
        self.metadata = _sa.MetaData()
        self.message_table = _sa.Table(
            self.message_table_name(name),
            self.metadata,
            _sa.Column("message_id", _UUIDType(), primary_key=True),
            _sa.Column(
                "global_position",
                _sa.Integer,
                _sa.Sequence(f"depeche_msgs_{name}_global_seq"),
                unique=True,
                nullable=False,
            ),
            _sa.Column(
                "added_at", _sa.DateTime, nullable=False, server_default=_sa.func.now()
            ),
            _sa.Column("stream", _sa.String(255), nullable=False),
            _sa.Column("version", _sa.Integer, nullable=False),
            _sa.Column("message", _PostgresJsonb, nullable=False),
            # TODO is this still required? only add in tests?
            _sa.UniqueConstraint(
                "stream", "version", name=f"depeche_msgs_{name}_version_uq"
            ),
        )
        self.notification_channel = self.notification_channel_name(name)
        ddl = _sa.DDL(
            "\n".join(
                [
                    _notify_trigger(
                        name=name,
                        tablename=self.message_table.name,
                        notification_channel=self.notification_channel,
                    ),
                    _write_message_fn(name=name, tablename=self.message_table.name),
                ]
            )
        )
        _sa.event.listen(
            self.message_table, "after_create", ddl.execute_if(dialect="postgresql")
        )
        self.metadata.create_all(engine, checkfirst=True)
        self._select = _sa.select(
            self.message_table.c.message_id,
            self.message_table.c.stream,
            self.message_table.c.version,
            self.message_table.c.message,
            self.message_table.c.global_position,
            self.message_table.c.added_at,
        )
        self._select_without_stream = _sa.select(
            self.message_table.c.message_id,
            self.message_table.c.version,
            self.message_table.c.message,
            self.message_table.c.global_position,
            self.message_table.c.added_at,
        )

    def add(
        self,
        conn: SAConnection,
        stream: str,
        expected_version: Optional[int],
        message_id: _uuid.UUID,
        message: dict,
    ) -> MessagePosition:
        return self.add_all(conn, stream, expected_version, [(message_id, message)])

    def add_all(
        self,
        conn: SAConnection,
        stream: str,
        expected_version: Optional[int],
        messages: Sequence[Tuple[_uuid.UUID, dict]],
    ) -> MessagePosition:
        assert len(messages) > 0
        func = getattr(_sa.func, f"depeche_write_message_{self.name}")
        try:
            for idx, (message_id, message) in enumerate(messages):
                _expected_version = (
                    expected_version + idx if expected_version is not None else None
                )
                result: Any = conn.execute(
                    _sa.select(
                        _sa.column("version"), _sa.column("global_position")
                    ).select_from(
                        func(
                            message_id, stream, PsycoPgJson(message), _expected_version
                        ).alias()
                    )
                )
        except _sa.exc.InternalError as exc:
            # psycopg2
            from depeche_db._compat import PsycoPgRaiseException

            if isinstance(exc.orig, PsycoPgRaiseException):
                raise OptimisticConcurrencyError(
                    f"optimistic concurrency failure: {exc.orig}"
                )
            raise
        except _sa.exc.ProgrammingError as exc:
            # psycopg3
            from depeche_db._compat import PsycoPgRaiseException

            if isinstance(exc.orig, PsycoPgRaiseException):
                raise OptimisticConcurrencyError(
                    f"optimistic concurrency failure: {exc.orig}"
                )
            raise
        row = result.fetchone()
        return MessagePosition(
            stream=stream,
            version=row.version,
            global_position=row.global_position,
        )

    def get_global_position(self, conn: SAConnection) -> int:
        result: Any = conn.execute(
            _sa.select(_sa.func.max(self.message_table.c.global_position))
        )
        return result.scalar() or 0

    def get_max_version(self, conn: SAConnection, stream: str) -> MessagePosition:
        row = conn.execute(
            _sa.select(
                _sa.func.max(self.message_table.c.version).label("version"),
                _sa.func.max(self.message_table.c.global_position).label(
                    "global_position"
                ),
            ).where(self.message_table.c.stream == stream),
        ).fetchone()
        if not row or row.version is None:
            return MessagePosition(stream, 0, None)
        return MessagePosition(stream, row.version, row.global_position)

    def get_message_ids(self, conn: SAConnection, stream: str) -> Iterator[_uuid.UUID]:
        for id in conn.execute(
            _sa.select(self.message_table.c.message_id)
            .where(self.message_table.c.stream == stream)
            .order_by(self.message_table.c.version)
        ).scalars():
            yield id

    def read(
        self, conn: SAConnection, stream: str
    ) -> Iterator[Tuple[_uuid.UUID, int, dict, int, _dt.datetime]]:
        return conn.execute(  # type: ignore
            self._select_without_stream.where(
                self.message_table.c.stream == stream
            ).order_by(self.message_table.c.version)
        )

    def read_multiple(
        self, conn: SAConnection, streams: Sequence[str]
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int, _dt.datetime]]:
        return conn.execute(  # type: ignore
            self._select.where(self.message_table.c.stream.in_(streams)).order_by(
                self.message_table.c.global_position
            )
        )

    def read_wildcard(
        self, conn: SAConnection, stream_wildcard: str
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int, _dt.datetime]]:
        return conn.execute(  # type: ignore
            self._select.where(
                self.message_table.c.stream.like(stream_wildcard)
            ).order_by(self.message_table.c.global_position)
        )

    def get_message_by_id(
        self, conn: SAConnection, message_id: _uuid.UUID
    ) -> Tuple[_uuid.UUID, str, int, dict, int, _dt.datetime]:
        return conn.execute(  # type: ignore
            self._select.where(self.message_table.c.message_id == message_id)
        ).first()

    def get_messages_by_ids(
        self, conn: SAConnection, message_ids: Sequence[_uuid.UUID]
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int, _dt.datetime]]:
        return conn.execute(  # type: ignore
            self._select.where(self.message_table.c.message_id.in_(message_ids))
        )

    def truncate(self, conn: SAConnection):
        conn.execute(self.message_table.delete())

    @staticmethod
    def message_table_name(name: str) -> str:
        return f"depeche_msgs_{name}"

    @staticmethod
    def notification_channel_name(name: str) -> str:
        return f"depeche_{name}_messages"

    @classmethod
    def get_migration_ddl(cls, name: str):
        """
        DDL Script to migrate from <=0.8.0
        """
        tablename = cls.message_table_name(name)
        new_objects = "\n".join(
            [
                _notify_trigger(
                    name=name,
                    tablename=tablename,
                    notification_channel=cls.notification_channel_name(name),
                ),
                _write_message_fn(name=name, tablename=tablename),
            ]
        )
        return f"""
            ALTER TABLE {name}_messages
                 RENAME TO {tablename};
            DROP TRIGGER {name}_notify_message_inserted;
            DROP FUNCTION IF EXISTS {name}_notify_message_inserted;
            DROP FUNCTION IF EXISTS {name}_write_message;
            {new_objects}
            """

    @classmethod
    def migrate_db_objects(cls, name: str, conn: SAConnection):
        """
        Migrate from <=0.8.0
        """
        conn.execute(cls.get_migration_ddl(name=name))


def _notify_trigger(name: str, tablename: str, notification_channel: str) -> str:
    trigger_name = f"depeche_storage_new_msg_{name}"
    return f"""
        CREATE OR REPLACE FUNCTION {trigger_name}()
          RETURNS trigger AS $$
        DECLARE
        BEGIN
          PERFORM pg_notify(
            '{notification_channel}',
            json_build_object(
                'message_id', NEW.message_id,
                'stream', NEW.stream,
                'version', NEW.version,
                'global_position', NEW.global_position
            )::text);
          RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;

        CREATE TRIGGER {trigger_name}
          AFTER INSERT ON {tablename}
          FOR EACH ROW
          EXECUTE PROCEDURE {trigger_name}();
     """


def _write_message_fn(name: str, tablename: str) -> str:
    function_name = f"depeche_write_message_{name}"
    return f"""
        CREATE OR REPLACE FUNCTION {function_name}(
          message_id uuid,
          stream varchar,
          message json,
          expected_version bigint DEFAULT NULL,
          OUT version bigint,
          OUT global_position bigint
        )
        AS $$
        DECLARE
          _stream_hash bigint;
          _stream_version bigint;
          _next_version bigint;
          _next_global_position bigint;
        BEGIN
          _stream_hash := left('x' || md5({function_name}.stream), 17)::bit(64)::bigint;
          PERFORM pg_advisory_xact_lock(_stream_hash);

          SELECT
            max({tablename}.version) into _stream_version
          FROM
            {tablename}
          WHERE
            {tablename}.stream = {function_name}.stream;

          IF _stream_version IS NULL THEN
            _stream_version := 0;
          END IF;

          IF {function_name}.expected_version IS NOT NULL THEN
            IF {function_name}.expected_version != _stream_version THEN
              RAISE EXCEPTION
                'Wrong expected version: %% (Stream: %%, Stream Version: %%)',
                {function_name}.expected_version,
                {function_name}.stream,
                _stream_version;
            END IF;
          END IF;

          _next_version := _stream_version + 1;
          _next_global_position := nextval('depeche_msgs_{name}_global_seq');

          INSERT INTO {tablename}
            (
              message_id,
              stream,
              version,
              message,
              global_position
            )
          VALUES
            (
              {function_name}.message_id,
              {function_name}.stream,
              _next_version,
              {function_name}.message,
              _next_global_position
            )
          ;

          version := _next_version;
          global_position := _next_global_position;
        END;
        $$ LANGUAGE plpgsql
        VOLATILE;
    """



from .db_lock_provider import DbLockProvider  # noqa: F401
from .db_subscription_state_provider import DbSubscriptionStateProvider  # noqa: F401
from .pg_notification_listener import PgNotificationListener  # noqa: F401
from .pydantic_message_serializer import PydanticMessageSerializer  # noqa: F401
from .stream_name_partitioner import StreamNamePartitioner  # noqa: F401



import inspect as _inspect
import typing as _typing
from typing import (
    Callable,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from ._compat import UNION_TYPES, issubclass_with_union
from ._interfaces import (
    HandlerDescriptor,
    MessageProtocol,
    StoredMessage,
    SubscriptionMessage,
)

E = TypeVar("E", bound=MessageProtocol)
H = TypeVar("H", bound=Callable)


class MessageHandlerRegister(Generic[E]):
    """
    Message handler register is a registry of message handlers.

    Typical usage:

        handlers = MessageHandlerRegister()

        @handlers.register
        def handle_message(message: MyMessage):
            ...

        @handlers.register
        def handle_other_message(message: StoredMessage[MyOtherMessage]):
            ...

    Implements: [MessageHandlerRegisterProtocol][depeche_db.MessageHandlerRegisterProtocol]
    """

    def __init__(
        self,
    ):
        self._handlers: Dict[Type[E], HandlerDescriptor] = {}

    def register(self, handler: H) -> H:
        """
        Registers a handler for a given message type.

        The handler must have at least one parameter. The first parameter must
        be of a message type. `E` being your message type, the parameter can be
        of type `E`, `SubscriptionMessage[E]` or `StoredMessage[E]`. When a
        handler is called, the message will be passed in the requested type.

        Multiple handlers can be registered for non-overlapping types of
        messages. Overlaps will cause a `ValueError`.

        Args:
            handler: A handler function.

        Returns:
            The unaltered handler function.
        """
        signature = _inspect.signature(handler)
        if len(signature.parameters) < 1:
            raise ValueError("Handler must have at least one parameter")

        handled_type = list(signature.parameters.values())[0].annotation
        self.register_manual(
            handler=handler,
            handled_type=handled_type,
            requires_middleware=len(signature.parameters) > 1,
        )
        return handler

    def register_manual(
        self, handler: H, handled_type: Type, requires_middleware: bool = False
    ):
        """
        Registers a handler for a given message type.

        The handler must have at least one parameter. The first parameter must
        be of a message type.

        Same overlap rules apply to this method as to the `register` method.

        If the handler takes more than one parameter, you must set the
        `requires_middleware` parameter to `True`!

        Args:
            handler: A handler function.
            handled_type: The type of message to handle.
            requires_middleware: Whether the handler requires middleware.
        """
        pass_subscription_message = False
        pass_stored_message = False
        origin = _typing.get_origin(handled_type)
        if origin == SubscriptionMessage:
            pass_subscription_message = True
            handled_type = handled_type.__args__[0]
        elif origin == StoredMessage:
            pass_stored_message = True
            handled_type = handled_type.__args__[0]

        if not issubclass_with_union(handled_type, MessageProtocol):
            raise TypeError(
                "Handled type (ie. the first argument) must be a subclass of MessageProtocol"
            )

        self.assert_not_registered(handled_type)
        self._handlers[handled_type] = HandlerDescriptor(
            handler=handler,
            pass_subscription_message=pass_subscription_message,
            pass_stored_message=pass_stored_message,
            requires_middleware=requires_middleware,
        )
        return handler

    def assert_not_registered(self, handled_type: Type[E]):
        if _typing.get_origin(handled_type) in UNION_TYPES:
            for member in _typing.get_args(handled_type):
                self.assert_not_registered(member)
        else:
            for registered_type in self._handlers:
                if issubclass_with_union(handled_type, registered_type):
                    raise ValueError(
                        f"Handler for {handled_type} is already registered for {registered_type}"
                    )

    def get_handler(self, message_type: Type[E]) -> Optional[HandlerDescriptor[E]]:
        for handled_type, handler in self._handlers.items():
            if issubclass_with_union(message_type, handled_type):
                return handler
        return None

    def get_all_handlers(self) -> _typing.Iterator[HandlerDescriptor[E]]:
        yield from self._handlers.values()


class MessageHandler(Generic[E]):
    """
    Message handler is a base class for message handlers.

    This is basically a class-based version of the `MessageHandlerRegister`.

    Typical usage (equivalent to the example in `MessageHandlerRegister`):

        class MyMessageHandler(MessageHandler):
            @MessageHandler.register
            def handle_message(self, message: MyMessage):
                ...

            @MessageHandler.register
            def handle_other_message(self, message: StoredMessage[MyOtherMessage]):
                ...

    Implements: [MessageHandlerRegisterProtocol][depeche_db.MessageHandlerRegisterProtocol]
    """

    _register: MessageHandlerRegister[E]

    def __init__(self):
        self._register = MessageHandlerRegister()
        for attrname in dir(self):
            attr = getattr(self, attrname)
            if getattr(attr, "__message_handler__", False):
                self._register.register(attr)

    @staticmethod
    def register(handler: H) -> H:
        handler.__message_handler__ = True  # type: ignore
        return handler

    def get_handler(self, message_type: Type[E]) -> Optional[HandlerDescriptor[E]]:
        return self._register.get_handler(message_type)

    def get_all_handlers(self) -> _typing.Iterator[HandlerDescriptor[E]]:
        yield from self._register.get_all_handlers()



from typing import AsyncGenerator, Generator, Generic, Iterable, Optional, TypeVar

import asyncer as _asyncer

from ._aggregated_stream import AggregatedStream
from ._interfaces import MessageProtocol, StoredMessage, SubscriptionStartPoint
from ._subscription import Subscription
from .tools import immem_subscription as _inmem_subscription
from .tools.async_pg_notification_listener import AsyncPgNotificationListener
from .tools.pg_notification_listener import PgNotificationListener

E = TypeVar("E", bound=MessageProtocol)


class AggregatedStreamReader(Generic[E]):
    def __init__(
        self,
        stream: AggregatedStream[E],
        start_point: Optional[SubscriptionStartPoint] = None,
    ):
        self.stream = stream
        self.subscription = stream.subscription(
            name=f"{stream.name}_reader",
            lock_provider=_inmem_subscription.NullLockProvider(),
            state_provider=_inmem_subscription.InMemorySubscriptionState(),
            start_point=start_point,
        )
        self.notification_listener = PgNotificationListener(
            dsn=stream._store.engine.url.render_as_string(hide_password=False),
            channels=[stream.notification_channel],
            select_timeout=0.5,
        )

    def start(self):
        """
        Start the notification listener.

        This method should be called before calling get_messages.

        The notification listener will listen for notifications on the
        notification channel of the stream. In order to do so, it creates a
        a connection to the database and listens for notifications on a separate
        thread. The connection is closed when the notification listener is stopped.
        """
        self.notification_listener.start()

    def get_messages(self, timeout: int = 0) -> Generator[StoredMessage[E], None, None]:
        """
        On the first call, get all messages in the stream after the start_point.
        On subsequent calls, get all messages in the stream after the last returned message.

        Args:
            timeout (int): Only wait for this many seconds when there are no new messages.
        """
        # Get messages already in the stream (after start_point)
        yield from _exhaust_subscription(self.subscription)

        # Get messages every time a notification is received
        for _ in self.notification_listener.messages(timeout=timeout):
            yield from _exhaust_subscription(self.subscription)

    def stop(self):
        """
        Stop the notification listener.
        """
        self.notification_listener.stop()


class AsyncAggregatedStreamReader(Generic[E]):
    def __init__(
        self,
        stream: AggregatedStream[E],
        start_point: Optional[SubscriptionStartPoint] = None,
    ):
        self.stream = stream
        self.subscription = stream.subscription(
            name=f"{stream.name}_reader",
            lock_provider=_inmem_subscription.NullLockProvider(),
            state_provider=_inmem_subscription.InMemorySubscriptionState(),
            start_point=start_point,
        )
        self.notification_listener = AsyncPgNotificationListener(
            dsn=stream._store.engine.url.render_as_string(hide_password=False),
            channels=[stream.notification_channel],
            select_timeout=0.5,
        )

    async def start(self):
        """
        Start the notification listener.
        """
        await self.notification_listener.start()

    async def get_messages(
        self, timeout: int = 0
    ) -> AsyncGenerator[StoredMessage[E], None]:
        """
        On the first call, get all messages in the stream after the start_point.
        On subsequent calls, get all messages in the stream after the last returned message.

        Args:
            timeout (int): Only wait for this many seconds when there are no new messages.
        """
        async_exhaust = _asyncer.asyncify(_exhaust_subscription)

        # Get messages already in the stream (after start_point)
        for message in await async_exhaust(self.subscription):
            yield message

        # Get messages every time a notification is received
        async for _ in self.notification_listener.messages(timeout=timeout):
            for message in await async_exhaust(self.subscription):
                yield message

    async def stop(self):
        """
        Stop the notification listener.
        """
        await self.notification_listener.stop()


def _exhaust_subscription(subscription: Subscription[E]) -> Iterable[StoredMessage[E]]:
    had_new_messages = True
    while had_new_messages:
        had_new_messages = False
        for message in subscription.get_next_messages(count=100):
            had_new_messages = True
            yield message.stored_message



import asyncio as _asyncio
import json as _json
import logging as _logging
import time
from typing import Any, AsyncIterator, Sequence

from .. import _compat
from .pg_notification_listener import PgNotification

logger = _logging.getLogger(__name__)


class AsyncPgNotificationListener:
    _conn: Any

    def __init__(
        self,
        dsn: str,
        channels: Sequence[str],
        select_timeout: float = 3.0,
        ignore_payload: bool = False,
    ):
        self.dsn = dsn
        self.channels = channels
        self._keep_running = True
        self._ignore_payload = ignore_payload
        self._select_timeout = select_timeout
        self._conn = None
        self._task = None

    async def start(self):
        """Start the notification listener."""
        assert (
            _compat.PSYCOPG3_AVAILABLE
        ), "AsyncPgNotificationListener requires psycopg3"
        import psycopg as _psycopg3

        dsn = self._parse_dsn(self.dsn)

        self._conn = await _psycopg3.AsyncConnection.connect(dsn)
        assert self._conn is not None

        async with self._conn.cursor() as cursor:
            for channel in self.channels:
                await cursor.execute(f"LISTEN {channel};")
            await self._conn.commit()

    async def messages(self, timeout: float = 0) -> AsyncIterator[PgNotification]:
        """Yield notifications as they arrive."""
        if self._conn is None:
            await self.start()
        assert self._conn is not None

        last_message_at = time.time()
        while self._keep_running:
            try:
                print("Waiting for notifications...", self.channels)
                async for notification in self._conn.notifies(
                    timeout=timeout if timeout > 0 else self._select_timeout
                ):
                    notification = await self._process_notification(notification)
                    last_message_at = time.time()
                    print("Async notification received:", notification)
                    if notification:
                        yield notification

                if timeout > 0:
                    if time.time() - last_message_at > timeout:
                        break
            except Exception as e:
                logger.exception(f"Error receiving notifications: {e}")
                await _asyncio.sleep(0.1)  # Prevent tight loop in case of errors

    async def stop(self):
        """Stop the notification listener and close the connection."""
        self._keep_running = False
        if self._conn:
            await self._conn.close()
            self._conn = None

    def _parse_dsn(self, dsn: str) -> str:
        import urllib.parse as _urlparse

        parts = list(_urlparse.urlparse(dsn))
        if parts[0] == "postgresql+psycopg":
            parts[0] = "postgresql"
        if parts[0] == "postgresql+psycopg2":
            parts[0] = "postgresql"
        dsn = _urlparse.urlunparse(parts)

        return dsn

    async def _process_notification(self, notification):
        try:
            if self._ignore_payload:
                payload = {}
            else:
                payload = _json.loads(notification.payload)
            return PgNotification(notification.channel, payload)
        except Exception:
            logger.exception(f"Error processing notification payload: {notification}")
            return None

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
